# -*- coding: utf-8 -*-
"""
core mixin module.
"""

import pyrin.utils.misc as misc_utils

from pyrin.core.structs import CoreObject, Hook
from pyrin.core.exceptions import InvalidHookTypeError


class HookMixin(CoreObject):
    """
    hook mixin class.

    every class that needs to provide hooks must inherit from this.
    """

    hook_type = Hook
    invalid_hook_type_error = InvalidHookTypeError

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of HookMixin.
        """

        super().__init__()

        self._hooks = []

    def _get_hooks(self):
        """
        gets all registered hooks.

        :returns: list[Hook]
        :rtype: list
        """

        return self._hooks

    def register_hook(self, instance):
        """
        registers the given instance into hooks.

        :param Hook instance: hook instance to be registered.

        :raises InvalidHookTypeError: invalid hook type error.
        """

        if not isinstance(instance, self.hook_type):
            full_name = misc_utils.try_get_fully_qualified_name(self.hook_type)
            raise self.invalid_hook_type_error('Input parameter [{instance}] is '
                                               'not an instance of [{hook}].'
                                               .format(instance=instance,
                                                       hook=full_name))
        self._hooks.append(instance)



# -*- coding: utf-8 -*-
"""
babel handlers package.
"""

from pyrin.packaging.base import Package


class BabelHandlersPackage(Package):
    """
    babel handlers package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.configuration']



# -*- coding: utf-8 -*-
"""
caching audit manager module.
"""

import traceback

import pyrin.caching.services as caching_services

from pyrin.audit.enumerations import InspectionStatusEnum
from pyrin.caching.audit import CachingAuditPackage
from pyrin.core.structs import Manager


class CachingAuditManager(Manager):
    """
    caching audit manager class.
    """

    package_class = CachingAuditPackage

    def inspect(self, **options):
        """
        inspects the status of available caches.

        it returns a tuple of two values. first value is a dict containing the inspection
        data. and the second value is a bool value indicating that inspection has been
        succeeded or failed.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :rtype: tuple[dict, bool]
        """

        all_caches = caching_services.get_cache_names()
        all_succeeded = True
        data = {}
        data.update(all_caches=all_caches)

        for name in all_caches:
            inspection_data, succeeded = self._test_cache(name, **options)
            if succeeded is False:
                all_succeeded = False

            data[name] = inspection_data

        return data, all_succeeded

    def _test_cache(self, name, **options):
        """
        tests that given cache is ready.

        it returns a tuple of two values. first value is a dict containing the inspection
        data. and the second value is a bool value indicating that inspection has been
        succeeded or failed.

        :param str name: cache name.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :rtype: tuple[dict, bool]
        """

        raise_error = options.get('raise_error', False)
        include_traceback = options.get('traceback', True)
        data = {}
        succeeded = True
        try:
            data.update(stats=caching_services.get_stats(name))
            key = 'audit.test'
            value = [1, 2, 3]
            caching_services.set(name, key, value, expire=5000)
            result = caching_services.get(name, key)
            caching_services.remove(name, key)

            if result != value:
                raise Exception('Result for key [{key}] did not returned '
                                'correctly from cache.'.format(key=key))

            data.update(status=InspectionStatusEnum.OK)
        except Exception as error:
            if raise_error is True:
                raise

            succeeded = False
            data.update(status=InspectionStatusEnum.FAILED,
                        error=str(error))
            if include_traceback is not False:
                data.update(traceback=traceback.format_exc())

        return data, succeeded



# -*- coding: utf-8 -*-
"""
schema result module.
"""

from copy import deepcopy

import pyrin.converters.serializer.services as serializer_services
import pyrin.configuration.services as config_services

from pyrin.core.globals import SECURE_TRUE, SECURE_FALSE
from pyrin.core.structs import CoreObject
from pyrin.api.schema.exceptions import SecureBooleanIsRequiredError, InvalidStartIndexError


class ResultSchema(CoreObject):
    """
    result schema class.
    """

    # these values could be set if you do not want to set them through '__init__' method.
    # read the '__init__' method's docstring for details.
    default_columns = None
    default_rename = None
    default_exclude = None
    default_depth = None
    default_readable = None
    default_indexed = None
    default_index_name = None
    default_start_index = None

    def __init__(self, **options):
        """
        initializes an instance of ResultSchema.

        note that at least one of keyword arguments must be provided.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided.
                                                      it will be used only for entity
                                                      conversion.

        :keyword dict[str, list[str]] | list[str] columns: column names to be included in result.
                                                           it could be a list of column names.
                                                           for example:
                                                           `columns=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity` and
                                                           `PersonEntity`, it should be like this:
                                                           `columns=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                           if provided column names are not
                                                           available in result, they will be
                                                           ignored.

        :note columns: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword dict[str, dict[str, str]] | dict[str, str] rename: column names that must be
                                                                    renamed in the result.
                                                                    it could be a dict with keys
                                                                    as original column names and
                                                                    values as new column names
                                                                    that should be exposed instead
                                                                    of original column names.
                                                                    for example:
                                                                    `rename=dict(age='new_age',
                                                                                 name='new_name')`
                                                                    but if you want to include
                                                                    relationships, then you must
                                                                    provide a dict containing
                                                                    entity class name as key and
                                                                    for value, another dict
                                                                    containing original column
                                                                    names as keys, and column
                                                                    names that must be exposed
                                                                    instead of original names,
                                                                    as values. for example
                                                                    if there is `CarEntity` and `
                                                                    PersonEntity`, it should be
                                                                    like this:
                                                                    `rename=
                                                                    dict(CarEntity=
                                                                         dict(name='new_name'),
                                                                         PersonEntity=
                                                                         dict(age='new_age')`
                                                                    then, the value of `name`
                                                                    column in result will be
                                                                    returned as `new_name` column.
                                                                    and also value of `age` column
                                                                    in result will be returned as
                                                                    'new_age' column. if provided
                                                                    rename columns are not
                                                                    available in result, they
                                                                    will be ignored.

        :note rename: dict[str entity_class_name, dict[str original_column, str new_column]] |
                      dict[str original_column, str new_column]

        :keyword dict[str, list[str]] | list[str] exclude: column names to be excluded from
                                                           result. it could be a list of column
                                                           names. for example:
                                                           `exclude=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity`
                                                           and `PersonEntity`, it should be
                                                           like this:
                                                           `exclude=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                            if provided excluded columns are not
                                                            available in result, they will be
                                                            ignored.

        :note exclude: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :raises SecureBooleanIsRequiredError: secure boolean is required error.
        :raises InvalidStartIndexError: invalid start index error.
        """

        super().__init__()

        columns = options.get('columns') or self.default_columns
        rename = options.get('rename') or self.default_rename
        exclude = options.get('exclude') or self.default_exclude
        depth = options.get('depth')
        readable = options.get('readable') or self.default_readable
        indexed = options.get('indexed')
        index_name = options.get('index_name') or self.default_index_name
        start_index = options.get('start_index')

        if depth is None:
            depth = self.default_depth

        if indexed is None:
            indexed = self.default_indexed

        if start_index is None:
            start_index = self.default_start_index

        self._default_index_name = config_services.get('api', 'schema', 'index_name')
        self._default_start_index = config_services.get('api', 'schema', 'start_index')

        self._columns = columns
        self._rename = rename
        self._exclude = exclude
        self._depth = depth
        self._indexed = indexed

        # set these to None to populate them using respective properties.
        self._readable = None
        self._index_name = None
        self._start_index = None

        # performing this through respective properties to get validations.
        self.readable = readable
        self.index_name = index_name
        self.start_index = start_index

    def _filter(self, item, **options):
        """
        filters the given item based on current schema.

        :param object item: item to be filtered.

        :keyword PaginatorBase paginator: paginator instance if any.
                                          if provided, it will be used to
                                          generate correct row indexes.

        :raises InvalidDepthProvidedError: invalid depth provided error.

        :returns: filtered object
        """

        if item is None:
            return item

        start_index = self.start_index
        paginator = options.get('paginator')
        if paginator is not None:
            start_index = (paginator.current_page
                           * paginator.current_page_size) - (paginator.current_page_size +
                                                             (-self.start_index))

        options.update(columns=self.columns,
                       rename=self.rename,
                       exclude=self.exclude,
                       depth=self.depth,
                       readable=self.readable,
                       indexed=self.indexed,
                       index_name=self.index_name,
                       start_index=start_index,
                       result_schema=self)

        return serializer_services.serialize(item, **options)

    def get_computed_row_columns(self, row, **options):
        """
        gets a dict containing all computed columns to be added to the result.

        this method is intended to be overridden in subclasses.
        note that the result dict should not contain any `BaseEntity` or
        `ROW_RESULT` values, otherwise a max recursion error may occur.

        :param ROW_RESULT row: the actual row result to be processed.

        :rtype: dict
        """

        return {}

    def get_computed_entity_columns(self, entity, **options):
        """
        gets a dict containing all computed columns to be added to the result.

        this method is intended to be overridden in subclasses.
        note that the result dict should not contain any `BaseEntity` or
        `ROW_RESULT` values, otherwise a max recursion error may occur.

        :param BaseEntity entity: the actual entity to be processed.

        :rtype: dict
        """

        return {}

    def copy(self):
        """
        returns a deep copy of this instance

        :rtype: ResultSchema
        """

        return deepcopy(self)

    def filter(self, item, **options):
        """
        filters the given item based on current schema.

        :param object item: item to be filtered.

        :keyword PaginatorBase paginator: paginator instance if any.
                                          if provided, it will be used to
                                          generate correct row indexes.

        :raises InvalidDepthProvidedError: invalid depth provided error.

        :returns: filtered object
        """

        return self._filter(item, **options)

    @property
    def columns(self):
        """
        gets the columns attribute of this schema.

        these are the columns that must be returned from the result for each record.

        :returns: dict[str entity_class_name, list[str column_name]] | list[str column_name]
        :rtype: dict[str, list[str]] | list[str]
        """

        return self._columns

    @property
    def rename(self):
        """
        gets the rename attribute of this schema.

        these are the columns that must be returned
        with modified names from the result for each record.

        :returns: dict[str entity_class_name, dict[str original_column, str new_column]] |
                  dict[str original_column, str new_column]

        :rtype: dict[str, dict[str, str]] | dict[str, str]
        """

        return self._rename

    @property
    def exclude(self):
        """
        gets the exclude attribute of this schema.

        these are the columns that must be excluded from the result for each record.

        :returns: dict[str entity_class_name, list[str column_name]] | list[str column_name]
        :rtype: dict[str, list[str]] | list[str]
        """

        return self._exclude

    @property
    def depth(self):
        """
        gets the depth attribute of this schema.

        this is a value that will be used to convert relationships.
        it will be used only for entity conversion.

        :rtype: int
        """

        return self._depth

    @depth.setter
    def depth(self, value):
        """
        sets the depth attribute of this schema.

        this is a value that will be used to convert relationships.
        it will be used only for entity conversion.

        :param int value: depth value to be set. it should not be higher
                          than `ConverterMixin.MAX_DEPTH` value.
        """

        self._depth = value

    @property
    def readable(self):
        """
        gets the readable attribute of this schema.

        it specifies that just readable attributes of an entity must be
        included in result. which are those that have `allow_read=True`
        and their name does not start with underscore `_`.
        it will be used only for entity conversion.

        :rtype: SECURE_TRUE | SECURE_FALSE
        """

        return self._readable

    @readable.setter
    def readable(self, value):
        """
        sets the readable attribute of this schema.

        it specifies that just readable attributes of an entity must be
        included in result. which are those that have `allow_read=True`
        and their name does not start with underscore `_`.
        it will be used only for entity conversion.

        :param SECURE_TRUE | SECURE_FALSE value: readable value to be set.

        :raises SecureBooleanIsRequiredError: secure boolean is required error.
        """

        if value not in (None, SECURE_TRUE, SECURE_FALSE):
            raise SecureBooleanIsRequiredError('The "readable" attribute of [{schema}] must '
                                               'be set to "SECURE_TRUE" or "SECURE_FALSE" but '
                                               'it is currently set to [{current}].'
                                               .format(schema=self, current=value))

        self._readable = value

    @property
    def indexed(self):
        """
        gets the indexed attribute of this schema.

        :rtype: bool
        """

        return self._indexed

    @indexed.setter
    def indexed(self, value):
        """
        sets the indexed attribute of this schema.

        :param bool value: specifies that this schema should return indexed results.
        """

        if value is None:
            value = False

        self._indexed = value

    @property
    def index_name(self):
        """
        gets the index name attribute of this schema.

        :rtype: str
        """

        return self._index_name

    @index_name.setter
    def index_name(self, value):
        """
        sets the index name attribute of this schema.

        :param str value: index field name.
        """

        if value is None:
            value = self._default_index_name

        self._index_name = value

    @property
    def start_index(self):
        """
        gets the start index attribute of this schema.

        :rtype: int
        """

        return self._start_index

    @start_index.setter
    def start_index(self, value):
        """
        sets the start index attribute of this schema.

        :param int value: start index.
        """

        if value is None:
            value = self._default_start_index

        if not isinstance(value, int):
            raise InvalidStartIndexError('The "start_index" attribute of [{schema}] '
                                         'must be set to an integer but it is currently '
                                         'set to [{current}].'
                                         .format(schema=self, current=value))
        self._start_index = value



# -*- coding: utf-8 -*-
"""
babel cli module.
"""

import pyrin.globalization.locale.babel.services as babel_services

from pyrin.cli.decorators import cli, cli_invoke, cli_group
from pyrin.core.structs import CLI


@cli_group('babel')
class BabelCLI(CLI):
    """
    babel cli class.

    this class exposes all babel cli commands.
    """

    _execute_service = babel_services.execute

    @cli_invoke
    def enable(self, include_pyrin=True, include_app=True, **options):
        """
        enables locale management for the application.

        :keyword bool include_pyrin: specifies that it should extract pyrin localizable
                                     messages. defaults to True if not provided.

        :keyword bool include_app: specifies that it should extract application
                                   localizable messages. defaults to True if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """

        return babel_services.enable(include_pyrin, include_app)

    @cli_invoke
    def rebuild(self, include_pyrin=True, include_app=True, **options):
        """
        it will do the three complete steps needed to
        update and compile locales with new messages.

        it will do:
            1. extract
            2. update
            3. compile

        this command is defined for convenient of usage, but if you need
        to do these steps separately, you could ignore this command and
        use the relevant command for each step.

        :keyword bool include_pyrin: specifies that it must extract pyrin localizable
                                     messages as well. defaults to True if not provided.

        :keyword bool include_app: specifies that it must extract application localizable
                                   messages as well. defaults to True if not provided.

        :keyword str locale: locale name of the catalog to compile.
                             it will compile all catalogs if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """

        return babel_services.rebuild(include_pyrin, include_app, **options)

    @cli
    def init(self, locale, **options):
        """
        create new message catalogs from a `.pot` file.

        :param str locale: locale name for the new localized catalog.
                           for example: `en` or `fr` or ...

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """

        return babel_services.check_init(locale)

    @cli
    def update(self, **options):
        """
        update existing message catalogs from a `.pot` file.

        :keyword str locale: locale name of the catalog to compile.
                             it will compile all catalogs if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def extract(self, include_pyrin=True, include_app=True, **options):
        """
        extract messages from source files and generate a `.pot` file.

        :keyword bool include_pyrin: specifies that it must extract pyrin localizable
                                     messages as well. defaults to True if not provided.

        :keyword bool include_app: specifies that it must extract application localizable
                                   messages as well. defaults to True if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def compile(self, **options):
        """
        compile message catalogs to `.mo` files.

        :keyword str locale: locale name of the catalog to compile.
                             it will compile all catalogs if not provided.

        :keyword bool statistics: print statistics about translations.
                                  defaults to False if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass



# -*- coding: utf-8 -*-
"""
logging masking component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.logging.masking import LoggingMaskingPackage
from pyrin.logging.masking.manager import LoggingMaskingManager


@component(LoggingMaskingPackage.COMPONENT_NAME)
class LoggingMaskingComponent(Component, LoggingMaskingManager):
    """
    logging masking component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli core template exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CLICoreTemplateManagerException(CoreException):
    """
    cli core template manager exception.
    """
    pass


class CLICoreTemplateManagerBusinessException(CoreBusinessException,
                                              CLICoreTemplateManagerException):
    """
    cli core template manager business exception.
    """
    pass


class InvalidProjectRootPathError(CLICoreTemplateManagerException):
    """
    invalid project root path error.
    """
    pass


class InvalidApplicationPackageNameError(CLICoreTemplateManagerException):
    """
    invalid application package name error.
    """
    pass


class InvalidApplicationClassNameError(CLICoreTemplateManagerException):
    """
    invalid application class name error.
    """
    pass


class InvalidProjectStructureTemplateHandlerType(CLICoreTemplateManagerException):
    """
    invalid project structure template handler type.
    """
    pass



# -*- coding: utf-8 -*-
"""
string normalizer handlers base module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.utilities.string.normalizer.interface import AbstractStringNormalizerBase
from pyrin.utilities.string.normalizer.handlers.exceptions import \
    InvalidStringNormalizerNameError, InvalidStringNormalizerPriorityError


class StringNormalizerBase(AbstractStringNormalizerBase):
    """
    string normalizer base class.
    """

    def __init__(self, name, priority, *args, **options):
        """
        initializes an instance of StringNormalizerBase.

        :param str name: name of this normalizer.
                         the normalizer will be registered by this name
                         into available normalizers. it must be unique.

        :param int priority: priority of this normalizer.
                             normalizers with higher priority will be executed sooner.

        :raises InvalidStringNormalizerNameError: invalid string normalizer name error.
        :raises InvalidStringNormalizerPriorityError: invalid string normalizer priority error.
        """

        super().__init__()

        if name in (None, '') or name.isspace():
            raise InvalidStringNormalizerNameError('String normalizer name must be '
                                                   'provided for [{normalizer}].'
                                                   .format(normalizer=self))

        if not isinstance(priority, int):
            raise InvalidStringNormalizerPriorityError('String normalizer [{normalizer}] '
                                                       'priority must be an integer.'
                                                       .format(normalizer=name))

        self._set_name(name)
        self._priority = priority

    def normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :keyword bool strip: strip spaces from both ends of value.
                             defaults to True if not provided.

        :returns: normalized value.
        :rtype: str
        """

        if value in (None, ''):
            return ''

        value = self._normalize(value, **options)
        strip = options.get('strip', True)
        if strip is not False:
            value = value.strip()

        return value

    @abstractmethod
    def _normalize(self, value, **options):
        """
        normalizes the given value.

        this method is intended to be overridden by subclasses.

        :param str value: value to be normalized.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: normalized value.
        :rtype: str
        """

        raise CoreNotImplementedError()

    @property
    def priority(self):
        """
        gets the priority of this normalizer.

        :rtype: int
        """

        return self._priority


class ReplaceNormalizerBase(StringNormalizerBase):
    """
    replace normalizer class.

    this normalizer replaces or removes provided values from string.
    """

    def __init__(self, name, priority, replace_map, **options):
        """
        initializes an instance of ReplaceNormalizerBase.

        :param str name: name of this normalizer.
                         the normalizer will be registered by this name
                         into available normalizers. it must be unique.

        :param int priority: priority of this normalizer.
                             normalizers with higher priority will be executed sooner.

        :param dict replace_map: a dict of keys and values to
                                 be used for replacement or removal.

        :raises InvalidStringNormalizerNameError: invalid string normalizer name error.
        :raises InvalidStringNormalizerPriorityError: invalid string normalizer priority error.
        """

        super().__init__(name, priority, **options)

        self._map = replace_map
        self._translation_table = str.maketrans(replace_map)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :returns: normalized value.
        :rtype: str
        """

        return value.translate(self._translation_table)



# -*- coding: utf-8 -*-
"""
utils exceptions module.
"""

from pyrin.core.exceptions import CoreException


class UtilsException(CoreException):
    """
    utils exception.
    """
    pass


class InputNotCallableError(UtilsException):
    """
    input not callable error.
    """
    pass


class ConfigurationFileNotFoundError(UtilsException):
    """
    configuration file not found error.
    """
    pass


class InvalidSchemaNameError(UtilsException):
    """
    invalid schema name error.
    """
    pass


class InvalidTableNameError(UtilsException):
    """
    invalid table name error.
    """
    pass


class InvalidColumnNameError(UtilsException):
    """
    invalid column name error.
    """
    pass


class InvalidRowResultFieldsAndValuesError(UtilsException):
    """
    invalid row result fields and values error.
    """
    pass


class FieldsAndValuesCountMismatchError(UtilsException):
    """
    fields and values count mismatch error.
    """
    pass


class DictKeyPrefixIsNotProvidedError(UtilsException):
    """
    dict key prefix is not provided error.
    """
    pass


class PathIsNotAbsoluteError(UtilsException):
    """
    path is not absolute error.
    """
    pass


class PathAlreadyExistedError(UtilsException):
    """
    path already existed error.
    """
    pass


class InvalidPathError(UtilsException):
    """
    invalid path error.
    """
    pass


class PathNotExistedError(UtilsException):
    """
    path not existed error.
    """
    pass


class IsNotDirectoryError(UtilsException):
    """
    is not directory error.
    """
    pass


class CheckConstraintValuesRequiredError(UtilsException):
    """
    check constraint values required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
schema exceptions module.
"""

from pyrin.core.exceptions import CoreException


class SchemaException(CoreException):
    """
    schema exception.
    """
    pass


class SecureBooleanIsRequiredError(SchemaException):
    """
    secure boolean is required error.
    """
    pass


class InvalidStartIndexError(SchemaException):
    """
    invalid start index error.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli manager module.
"""

import colorama

import pyrin.utils.function as func_utils

from pyrin.cli import CLIPackage
from pyrin.cli.structs import CLIGroups
from pyrin.core.structs import Manager, CLI
from pyrin.utils.custom_print import print_colorful, print_error
from pyrin.cli.exceptions import InvalidCLIDecoratedMethodError, InvalidCLIGroupTypeError, \
    CLIGroupNameIsRequiredError, DuplicatedCLIGroupError


class CLIManager(Manager):
    """
    cli manager class.
    """

    package_class = CLIPackage

    def __init__(self):
        """
        initializes an instance of CLIManager.
        """

        super().__init__()

        self.__groups = CLIGroups()

    def register_cli_group(self, name, instance):
        """
        registers a cli group.

        :param str name: cli group name.
        :param CLI instance: cli group instance.

        :raises CLIGroupNameIsRequiredError: cli group name is required error.
        :raises DuplicatedCLIGroupError: duplicated cli group error.
        :raises InvalidCLIGroupTypeError: invalid cli group type error.
        """

        if name in (None, '') or name.isspace():
            raise CLIGroupNameIsRequiredError('CLI group name is required for '
                                              'registering cli group [{instance}].'
                                              .format(instance=instance))

        if hasattr(self.__groups, name) is True:
            raise DuplicatedCLIGroupError('There is another registered cli group with '
                                          'name [{name}]. so cli group [{instance}] '
                                          'could not be registered.'
                                          .format(name=name, instance=instance))

        if not isinstance(instance, CLI):
            raise InvalidCLIGroupTypeError('CLI group [{instance}] is not '
                                           'an instance of [{base}].'
                                           .format(instance=instance, base=CLI))

        setattr(self.__groups, name, instance)

    def get_cli_groups(self):
        """
        gets all registered cli groups.

        :rtype: CLIGroups
        """

        return self.__groups

    def process_function(self, func, func_args, func_kwargs):
        """
        processes the given cli handler function with given inputs.

        :param function func: function to to be executed.
        :param tuple func_args: a tuple of function positional inputs.
        :param dict func_kwargs: a dictionary of function keyword arguments.

        :raises CLIHandlerNotFoundError: cli handler not found error.
        :raises InvalidCLIDecoratedMethodError: invalid cli decorated method error.

        :rtype: int
        """

        try:
            inputs, cli_instance = func_utils.get_inputs(func, func_args, func_kwargs)
            func_kwargs.pop('help', None)
            if cli_instance is None:
                raise InvalidCLIDecoratedMethodError('The "@cli" decorator must '
                                                     'be set on instance methods. '
                                                     'static methods or stand-alone '
                                                     'functions are not valid.')

            if self._process_help(func, inputs) is False:
                func(*func_args, **func_kwargs)
                return cli_instance.execute(func.__name__, **inputs)

        except TypeError as error:
            print_error(str(error) + '\n', force=True)
            self._print_function_doc(func)

    def invoke_function(self, func, func_args, func_kwargs):
        """
        invokes the given function with given inputs.

        :param function func: function to be executed.
        :param tuple func_args: a tuple of function positional inputs.
        :param dict func_kwargs: a dictionary of function keyword arguments.

        :returns: function execution result.
        """

        try:
            inputs, parent = func_utils.get_inputs(func, func_args, func_kwargs)
            func_kwargs.pop('help', None)
            if self._process_help(func, inputs) is False:
                return func(*func_args, **func_kwargs)

        except TypeError as error:
            print_error(str(error) + '\n', force=True)
            self._print_function_doc(func)

    def _print_function_doc(self, func):
        """
        prints the docstring of given function.

        :param function func: function to print its docstring.
        """

        doc = func_utils.get_doc(func, False)
        if doc is not None and len(doc) > 0:
            result = '`{func}` command usage:\n\neach argument could be passed ' \
                     'with `--arg value` format\nor all arguments could be passed in ' \
                     'positional order.\nfor boolean arguments, you just need to pass ' \
                     '`--arg` without any \nvalue for True and do not pass it to use ' \
                     'default value (often False).\n\n`{func}` command doc:\n\n{doc}' \
                     .format(func=func.__name__, doc=doc)
        else:
            result = '`{func}` command help is not available.'.format(func=func.__name__)

        print_colorful(result, colorama.Fore.CYAN, force=True)

    def _process_help(self, func, inputs):
        """
        processes the function to print help message if required.

        :param function func: function to process help for it.
        :param dict inputs: a dict of function inputs.
        """

        if inputs.pop('help', False) is True:
            self._print_function_doc(func)
            return True

        return False



# -*- coding: utf-8 -*-
"""
orm query exceptions module.
"""

from pyrin.core.exceptions import CoreException


class ORMQueryException(CoreException):
    """
    orm query exception.
    """
    pass


class ORMQueryBusinessException(ORMQueryException):
    """
    orm query business exception.
    """
    pass


class ColumnsOutOfScopeError(ORMQueryBusinessException):
    """
    columns out of scope error.
    """
    pass


class UnsupportedQueryStyleError(ORMQueryException):
    """
    unsupported query style error.
    """
    pass


class InjectTotalCountError(ORMQueryException):
    """
    inject total count error.
    """
    pass



# -*- coding: utf-8 -*-
"""
packaging services module.
"""

from pyrin.packaging import PackagingPackage
from pyrin.application.services import get_component


def load_components(**options):
    """
    loads required packages and modules for application startup.

    :raises BothUnitAndIntegrationTestsCouldNotBeLoadedError: both unit and integration
                                                              tests could not be loaded
                                                              error.
    :raises PackageIsIgnoredError: package is ignored error.
    :raises PackageIsDisabledError: package is disabled error.
    :raises PackageNotExistedError: package not existed error.
    :raises SelfDependencyDetectedError: self dependency detected error.
    :raises SubPackageDependencyDetectedError: sub-package dependency detected error.
    :raises CircularDependencyDetectedError: circular dependency detected error.
    :raises PackageExternalDependencyError: package external dependency error.
    """

    get_component(PackagingPackage.COMPONENT_NAME).load_components(**options)


def load(module_name, **options):
    """
    loads the specified module.

    :param str module_name: full module name.
                            example module_name = `pyrin.application.decorators`.

    :rtype: Module
    """

    return get_component(PackagingPackage.COMPONENT_NAME).load(module_name, **options)


def register_hook(instance):
    """
    registers the given instance into packaging hooks.

    :param PackagingHookBase instance: packaging hook instance to be registered.

    :raises InvalidPackagingHookTypeError: invalid packaging hook type error.
    """

    get_component(PackagingPackage.COMPONENT_NAME).register_hook(instance)


def get_working_directory(root_path):
    """
    gets working directory path according to given root path.

    working directory is where the root application and test package are resided.
    this is required when application starts from any of test applications.
    then we should move root path up, to the correct root to be able to
    include real application packages too.
    if the application has been started from real application, this method
    returns the same input.

    :param str root_path: root path to get working directory from.

    :rtype: str
    """

    return get_component(PackagingPackage.COMPONENT_NAME).get_working_directory(root_path)


def get_loaded_packages():
    """
    gets the name of all loaded packages.

    :rtype: list[str]
    """

    return get_component(PackagingPackage.COMPONENT_NAME).get_loaded_packages()


def is_package_loaded(name):
    """
    gets a value indicating that given package is loaded.

    :param str name: package fully qualified name.

    :rtype: bool
    """

    return get_component(PackagingPackage.COMPONENT_NAME).is_package_loaded(name)



# -*- coding: utf-8 -*-
"""
template handlers package.
"""

from pyrin.packaging.base import Package


class TemplateHandlersPackage(Package):
    """
    template handlers package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
model package.
"""

from pyrin.packaging.base import Package


class ModelPackage(Package):
    """
    model package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.model.component'



# -*- coding: utf-8 -*-
"""
core globals module.
"""

from flask_babel import gettext
from sqlalchemy.util._collections import AbstractKeyedTuple


# this value should be used as `None`, where the `None` itself has a meaning.
NULL = object()

# this value should be used where we need to reference class type of None objects.
NONE_TYPE = type(None)

# this value should be used where we want to determine if an object is a list type or not.
LIST_TYPES = (list, tuple, set)

# this method should be used for localizable strings.
_ = gettext

# this value should be used when working on AbstractKeyedTuple objects.
ROW_RESULT = AbstractKeyedTuple

# this value should be used instead of True when we want to preserve security of
# inputs which come from client to prevent injection of invalid values through options.
SECURE_TRUE = object()

# this value should be used instead of False when we want to preserve security of
# inputs which come from client to prevent injection of invalid values through options.
SECURE_FALSE = object()



# -*- coding: utf-8 -*-
"""
request wrappers exceptions module.
"""

from pyrin.core.enumerations import ClientErrorResponseCodeEnum
from pyrin.core.exceptions import CoreException, CoreBusinessException


class RequestWrappersException(CoreException):
    """
    request wrappers exception.
    """
    pass


class RequestWrappersBusinessException(CoreBusinessException,
                                       RequestWrappersException):
    """
    request wrappers business exception.
    """
    pass


class InvalidRequestContextKeyNameError(RequestWrappersException):
    """
    invalid request context key name error.
    """
    pass


class RequestContextKeyIsAlreadyPresentError(RequestWrappersException):
    """
    request context key is already present error.
    """
    pass


class BadRequestError(RequestWrappersBusinessException):
    """
    bad request error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of BadRequestError.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.BAD_REQUEST


class JSONBodyDecodingError(BadRequestError):
    """
    json body decoding error.
    """
    pass


class BodyDecodingError(BadRequestError):
    """
    body decoding error.
    """
    pass


class RequestDeserializationError(BadRequestError):
    """
    request deserialization error.
    """
    pass


class RequestComponentCustomKeyAlreadySetError(RequestWrappersException):
    """
    request component custom key already set error.
    """
    pass


class LargeContentError(RequestWrappersBusinessException):
    """
    large content error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of LargeContentError.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.PAYLOAD_TOO_LARGE



# -*- coding: utf-8 -*-
"""
database paging package.
"""

from pyrin.packaging.base import Package


class DatabasePagingPackage(Package):
    """
    database paging package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.paging.component'
    DEPENDS = ['pyrin.configuration']



# -*- coding: utf-8 -*-
"""
deserializer services module.
"""

from pyrin.converters.deserializer import DeserializerPackage
from pyrin.application.services import get_component


def deserialize(value, **options):
    """
    deserializes the given value.

    returns deserialized object on success or returns
    the same input value if deserialization fails.

    :param object value: value to be deserialized.

    :keyword bool include_internal: specifies that any chained internal deserializer
                                    must also be used for deserialization. if set to
                                    False, only non-internal deserializers will be used.
                                    defaults to True if not provided.

    :returns: deserialized object
    """

    return get_component(DeserializerPackage.COMPONENT_NAME).deserialize(value, **options)


def register_deserializer(instance, **options):
    """
    registers a new deserializer or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a deserializer which is already registered.

    :param AbstractDeserializerBase instance: deserializer to be registered.
                                              it must be an instance of
                                              AbstractDeserializerBase.

    :keyword bool replace: specifies that if there is another registered
                           deserializer with the same name and accepted type,
                           replace it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidDeserializerTypeError: invalid deserializer type error.
    :raises DuplicatedDeserializerError: duplicated deserializer error.
    """

    return get_component(DeserializerPackage.COMPONENT_NAME).register_deserializer(instance,
                                                                                   **options)


def get_deserializers(**options):
    """
    gets all registered deserializers.

    it could filter deserializers for a specific type if provided.
    it only returns the first deserializer for each type, because
    all deserializers for a given type, are chained together.

    :keyword type accepted_type: specifies to get deserializers which are registered for
                                 the accepted type. if not provided, all deserializers
                                 will be returned.

    :rtype: list[AbstractDeserializerBase]
    """

    return get_component(DeserializerPackage.COMPONENT_NAME).get_deserializers(**options)



# -*- coding: utf-8 -*-
"""
deserializer interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class DeserializerSingletonMeta(MultiSingletonMeta):
    """
    deserializer singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractDeserializerBase(CoreObject, metaclass=DeserializerSingletonMeta):
    """
    abstract deserializer base class.
    """

    @abstractmethod
    def deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.

        :param object value: value to be deserialized.

        :keyword bool include_internal: specifies that any chained internal deserializer
                                        must also be used for deserialization. if set to
                                        False, only non-internal deserializers will be used.
                                        defaults to True if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: deserialized value.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def set_next(self, deserializer):
        """
        sets the next deserializer handler and returns it.

        :param AbstractDeserializerBase deserializer: deserializer instance to
                                                      be set as next handler.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: AbstractDeserializerBase
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def accepted_type(self):
        """
        gets the accepted type for this deserializer.

        which could deserialize values from this type.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def internal(self):
        """
        gets a value indicating that this deserializer is internal.

        internal deserializers will not be used for deserializing client inputs.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
logging masking structs module.
"""

from werkzeug.datastructures import Headers

import pyrin.logging.masking.services as masking_services
import pyrin.utils.headers as header_utils

from pyrin.core.globals import LIST_TYPES
from pyrin.core.structs import CoreImmutableDict


class MaskedDict(CoreImmutableDict):
    """
    masked dict class.

    this class is a normal dict with the ability of masking
    values of specific keys in the dict.
    note that this type is immutable. meaning that values or keys of
    dict could not be changed after initialization.
    """

    # the value to be used for masked keys.
    MASK = '*'

    def __init__(self, mapping=None, **kwargs):
        """
        initializes an instance of MaskedDict.

        :param dict | MaskedDict | Headers | iterable mapping: values to create a
                                                               dict form them.
        """

        masked_kwargs = {}
        if len(kwargs) > 0:
            masked_kwargs = self._mask_dict(kwargs)

        if isinstance(mapping, MaskedDict):
            super().__init__(self._unify(mapping, masked_kwargs))

        elif isinstance(mapping, (dict, Headers)):
            temp = {}
            for key, value in mapping.items():
                if masking_services.should_mask(key) is True:
                    temp[key] = self.MASK
                else:
                    temp[key] = self._mask(value)

            super().__init__(self._unify(temp, masked_kwargs))

        elif isinstance(mapping, LIST_TYPES):
            temp = []
            for key, value in mapping:
                if masking_services.should_mask(key) is True:
                    temp.append((key, self._mask(value)))
                else:
                    temp.append((key, value))

            super().__init__(temp, **masked_kwargs)

        else:
            super().__init__(masked_kwargs)

    def _unify(self, mapping, kwargs):
        """
        returns a unified dict from given mapping dict and keyword arguments dict.

        :param dict mapping: a mapping dict.
        :param dict kwargs: an extra dict to be unified with mapping.

        :rtype: dict
        """

        if len(kwargs) <= 0:
            return mapping

        data = dict(mapping)
        data.update(kwargs)
        return data

    def _mask(self, value):
        """
        masks all required keys of given object.

        if given object is not a dict or Headers or a collection of those
        objects, it returns the same input value.

        :param dict | Headers | list | tuple | set value: a dict or Headers or a
                                                          collection of those types.

        :rtype: dict | list[dict] | tuple[dict] | set[dict]
        """

        if isinstance(value, Headers):
            return self._mask_headers(value)
        elif isinstance(value, dict):
            return self._mask_dict(value)
        elif isinstance(value, LIST_TYPES):
            return self._mask_list(value)
        return value

    def _mask_dict(self, value):
        """
        masks all required keys of given dict.

        :param dict value: dict to mask its keys.

        :rtype: dict
        """

        return MaskedDict(value)

    def _mask_headers(self, value):
        """
        masks all required keys of given Headers object.

        :param Headers value: Headers to mask its keys.

        :rtype: dict
        """

        headers = header_utils.convert_headers(value)
        return self._mask_dict(headers.to_dict())

    def _mask_list(self, items):
        """
        masks all required keys of different items of given iterable.

        :param iterable items: items to mask their keys.

        :rtype: list | tuple | set
        """

        result_type = type(items)
        results = []
        for item in items:
            results.append(self._mask(item))

        return result_type(results)



# -*- coding: utf-8 -*-
"""
validator manager module.
"""

from pyrin.core.globals import _
from pyrin.core.structs import Manager, Context, DTO
from pyrin.utils.custom_print import print_warning
from pyrin.validator import ValidatorPackage
from pyrin.validator.interface import AbstractValidatorBase
from pyrin.validator.exceptions import InvalidValidatorTypeError, DuplicatedValidatorError, \
    ValidatorNotFoundError, ValidationError, InvalidEntityForValidationError, \
    InvalidDataForValidationError, ValidatorDomainNotFoundError


class ValidatorManager(Manager):
    """
    validator manager class.
    """

    package_class = ValidatorPackage

    def __init__(self):
        """
        initializes an instance of ValidatorManager.
        """

        super().__init__()

        # a dictionary containing information of registered validators.
        # example: dict(type[BaseEntity] |
        #               str domain: dict(str name: AbstractValidatorBase instance))
        self._validators = Context()

    def register_validator(self, instance, **options):
        """
        registers a new validator or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a validator which is already registered.

        :param AbstractValidatorBase instance: validator to be registered.
                                               it must be an instance of
                                               AbstractValidatorBase.

        :keyword bool replace: specifies that if there is another registered
                               validator with the same domain and name, replace
                               it with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidValidatorTypeError: invalid validator type error.
        :raises DuplicatedValidatorError: duplicated validator error.
        """

        if not isinstance(instance, AbstractValidatorBase):
            raise InvalidValidatorTypeError('Input parameter [{instance}] is '
                                            'not an instance of [{base}].'
                                            .format(instance=instance,
                                                    base=AbstractValidatorBase))

        domain_validators = self._validators.get(instance.domain)
        if domain_validators is not None:
            old_instance = domain_validators.get(instance.name)
            if old_instance is not None:
                replace = options.get('replace', False)
                if replace is not True:
                    raise DuplicatedValidatorError('There is another registered '
                                                   'validator [{old}] with name '
                                                   '[{name}] for domain [{domain}] '
                                                   'but "replace" option is not set, '
                                                   'so validator [{instance}] '
                                                   'could not be registered.'
                                                   .format(old=old_instance,
                                                           name=instance.name,
                                                           domain=instance.domain,
                                                           instance=instance))

                print_warning('Validator [{old_instance}] is going '
                              'to be replaced by [{new_instance}] '
                              'for domain [{domain}].'
                              .format(old_instance=old_instance,
                                      new_instance=instance,
                                      domain=instance.domain))

        if domain_validators is None:
            domain_validators = DTO()
            self._validators[instance.domain] = domain_validators
        domain_validators[instance.name] = instance
        self._validators[instance.domain] = domain_validators

    def get_domain_validators(self, domain):
        """
        gets all registered validators for given domain.

        it raises an error if domain does not exist.

        :param type[BaseEntity] | str domain: the domain to get its validators.
                                              it could be a type of a BaseEntity
                                              subclass or a string name.

        :raises ValidatorDomainNotFoundError: validator domain not found error.

        :rtype: dict[type[BaseEntity] | str, AbstractValidatorBase]
        """

        if domain not in self._validators:
            raise ValidatorDomainNotFoundError('Validator domain [{name}] does not exist.'
                                               .format(name=domain))

        return self._validators.get(domain)

    def get_validator(self, domain, name):
        """
        gets the registered validator for given domain and name.

        it returns None if no validator found for given name.

        :param type[BaseEntity] | str domain: the domain to get validator from.
                                              it could be a type of a BaseEntity
                                              subclass or a string name.

        :param str name: validator name to get.

        :raises ValidatorDomainNotFoundError: validator domain not found error.

        :rtype: AbstractValidatorBase
        """

        domain_validators = self.get_domain_validators(domain)
        return domain_validators.get(name)

    def validate_field(self, domain, name, value, **options):
        """
        validates the given value with given validator.

        it returns the same value or fixed one.

        :param type[BaseEntity] | str domain: the domain to validate the value for.
                                              it could be a type of a BaseEntity
                                              subclass or a string name.

        :param str name: validator name to be used for validation.
        :param object | list[object] value: value to be validated.

        :keyword bool force: specifies that if there is no validator
                             with specified domain and name, it should
                             raise an error. defaults to False if not provided.

        :keyword bool nullable: determines that provided value could be None.
        :keyword bool is_list: specifies that the value must be a list of items.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this argument will only be considered in min
                                         and range validators.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this argument will only be considered in max
                                         and range validators.

        :keyword bool allow_blank: determines that empty strings should be
                                   considered valid. this argument will only
                                   be considered in string validators.

        :keyword bool allow_whitespace: determines that whitespace strings should be
                                        considered valid. this argument will only
                                        be considered in string validators.

        :raises ValidatorDomainNotFoundError: validator domain not found error.
        :raises ValidatorNotFoundError: validator not found error.
        :raises ValidationError: validation error.

        :returns: same value or fixed one.
        """

        validator = self.get_validator(domain, name)
        force = options.get('force')
        if force is None:
            force = False

        if force is True and validator is None:
            raise ValidatorNotFoundError('There is no validator with name '
                                         '[{name}] for domain [{domain}].'
                                         .format(name=name, domain=domain))

        if validator is not None:
            return validator.validate(value, **options)

        return value

    def validate_dict(self, domain, data, **options):
        """
        validates available values of given dict.

        it uses the correct validator for each value based on its key name.

        :param type[BaseEntity] | str domain: the domain to validate the values for.
                                              it could be a type of a BaseEntity
                                              subclass or a string name.

        :param dict data: dictionary to validate its values.

        :keyword bool lazy: specifies that all values must be validated first and
                            then a cumulative error must be raised containing a dict
                            of all keys and their corresponding error messages.
                            defaults to True if not provided.

        :keyword bool for_update: specifies that only fields that are present in data
                                  must be validated. even if there are validators for
                                  other fields. it is useful for changing the validation
                                  behavior on insert or update operations. defaults to
                                  False if not provided and all validators will be used.

        :keyword bool nullable: determines that provided values could be None.
        :keyword bool is_list: specifies that the value must be a list of items.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool allow_blank: determines that empty strings should be
                                   considered valid. this argument will only
                                   be considered in string validators.

        :keyword bool allow_whitespace: determines that whitespace strings should be
                                        considered valid. this argument will only
                                        be considered in string validators.

        :keyword BaseEntity entity: an entity instance that the provided data
                                    is the result dict of it.
                                    it will be used to populate fixed values
                                    in the entity.

        :raises InvalidDataForValidationError: invalid data for validation error.
        :raises ValidatorDomainNotFoundError: validator domain not found error.
        :raises ValidatorNotFoundError: validator not found error.
        :raises ValidationError: validation error.
        """

        if data is None:
            raise InvalidDataForValidationError(_('Data for validation could not be None.'))

        cumulative_errors = DTO()
        lazy = options.get('lazy', True)
        for_update = options.get('for_update', False)
        entity = options.get('entity')

        available_data = set(data.keys())
        validator_names = set()
        available_validators = self.get_domain_validators(domain)
        if available_validators is not None:
            validator_names = set(available_validators.keys())

        should_be_validated = validator_names
        if for_update is True:
            should_be_validated = validator_names.intersection(available_data)

        for name in should_be_validated:
            try:
                fixed_value = self.validate_field(domain, name, data.get(name), **options)
                if fixed_value is not None:
                    data[name] = fixed_value
                    if entity is not None:
                        entity.set_attribute(name, fixed_value, silent=True)

            except ValidationError as error:
                if lazy is False:
                    raise error
                else:
                    cumulative_errors[name] = error.description

        if len(cumulative_errors) > 0:
            raise ValidationError(_('Validation failed with some errors.'),
                                  data=cumulative_errors)

    def validate_entity(self, entity, **options):
        """
        validates available values of given entity.

        it uses the correct validator for each value based on its field name.

        :param BaseEntity entity: entity to validate its values.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in validation. defaults to
                                                      `SECURE_TRUE` if not provided.

        :keyword bool lazy: specifies that all fields must be validated first and
                            then a cumulative error must be raised containing a dict
                            of all field names and their corresponding error messages.
                            defaults to True if not provided.

        :keyword bool for_update: specifies that only fields that are present in data
                                  must be validated. even if there are validators for
                                  other fields. it is useful for changing the validation
                                  behavior on insert or update operations. defaults to
                                  False if not provided and all validators will be used.

        :keyword bool nullable: determines that provided values could be None.
        :keyword bool is_list: specifies that the value must be a list of items.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool allow_blank: determines that empty strings should be
                                   considered valid. this argument will only
                                   be considered in string validators.

        :keyword bool allow_whitespace: determines that whitespace strings should be
                                        considered valid. this argument will only
                                        be considered in string validators.

        :raises InvalidEntityForValidationError: invalid entity for validation error.
        :raises ValidatorDomainNotFoundError: validator domain not found error.
        :raises ValidatorNotFoundError: validator not found error.
        :raises ValidationError: validation error.
        """

        if entity is None:
            raise InvalidEntityForValidationError(_('Entity for validation could not be None.'))

        options.update(entity=entity)
        self.validate_dict(type(entity), entity.to_dict(**options), **options)

    def is_valid_field(self, domain, name, value, **options):
        """
        gets a value indicating that given field is valid.

        :param type[BaseEntity] | str domain: the domain to validate the value for.
                                              it could be a type of a BaseEntity
                                              subclass or a string name.

        :param str name: validator name to be used for validation.
        :param object | list[object] value: value to be validated.

        :keyword bool force: specifies that if there is no validator
                             with specified domain and name, it should
                             raise an error. defaults to False if not provided.

        :keyword bool nullable: determines that provided value could be None.
        :keyword bool is_list: specifies that the value must be a list of items.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this argument will only be considered in min
                                         and range validators.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this argument will only be considered in max
                                         and range validators.

        :keyword bool allow_blank: determines that empty strings should be
                                   considered valid. this argument will only
                                   be considered in string validators.

        :keyword bool allow_whitespace: determines that whitespace strings should be
                                        considered valid. this argument will only
                                        be considered in string validators.

        :raises ValidatorDomainNotFoundError: validator domain not found error.
        :raises ValidatorNotFoundError: validator not found error.

        :rtype: bool
        """

        try:
            fixed_value = self.validate_field(domain, name, value, **options)
            return value == fixed_value
        except ValidationError:
            return False

    def is_valid_dict(self, domain, data, **options):
        """
        gets a value indicating that given dict has valid values.

        it uses the correct validator for each value based on its key name.

        :param type[BaseEntity] | str domain: the domain to validate the values for.
                                              it could be a type of a BaseEntity
                                              subclass or a string name.

        :param dict data: dictionary to validate its values.

        :keyword bool for_update: specifies that only fields that are present in data
                                  must be validated. even if there are validators for
                                  other fields. it is useful for changing the validation
                                  behavior on insert or update operations. defaults to
                                  False if not provided and all validators will be used.

        :keyword bool nullable: determines that provided values could be None.
        :keyword bool is_list: specifies that the value must be a list of items.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool allow_blank: determines that empty strings should be
                                   considered valid. this argument will only
                                   be considered in string validators.

        :keyword bool allow_whitespace: determines that whitespace strings should be
                                        considered valid. this argument will only
                                        be considered in string validators.

        :raises InvalidDataForValidationError: invalid data for validation error.
        :raises ValidatorDomainNotFoundError: validator domain not found error.
        :raises ValidatorNotFoundError: validator not found error.

        :rtype: bool
        """

        try:
            options.update(lazy=False)
            self.validate_dict(domain, data, **options)
            return True
        except ValidationError:
            return False

    def is_valid_entity(self, entity, **options):
        """
        gets a value indicating that given entity has valid values.

        it uses the correct validator for each value based on its field name.

        :param BaseEntity entity: entity to validate its values.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in validation. defaults to
                                                      `SECURE_TRUE` if not provided.

        :keyword bool for_update: specifies that only fields that are present in data
                                  must be validated. even if there are validators for
                                  other fields. it is useful for changing the validation
                                  behavior on insert or update operations. defaults to
                                  False if not provided and all validators will be used.

        :keyword bool nullable: determines that provided values could be None.
        :keyword bool is_list: specifies that the value must be a list of items.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this argument will only be considered in min,
                                         max and range validators.

        :keyword bool allow_blank: determines that empty strings should be
                                   considered valid. this argument will only
                                   be considered in string validators.

        :keyword bool allow_whitespace: determines that whitespace strings should be
                                        considered valid. this argument will only
                                        be considered in string validators.

        :raises InvalidEntityForValidationError: invalid entity for validation error.
        :raises ValidatorDomainNotFoundError: validator domain not found error.
        :raises ValidatorNotFoundError: validator not found error.

        :rtype: bool
        """

        try:
            options.update(lazy=False)
            self.validate_entity(entity, **options)
            return True
        except ValidationError:
            return False



# -*- coding: utf-8 -*-
"""
database paging exceptions module.
"""

from pyrin.core.exceptions import CoreException


class DatabasePagingException(CoreException):
    """
    database paging exception.
    """
    pass


class DatabasePagingBusinessException(DatabasePagingException):
    """
    database paging business exception.
    """
    pass


class PageSizeLimitError(DatabasePagingException):
    """
    page size limit error.
    """
    pass


class TotalCountIsAlreadySetError(DatabasePagingException):
    """
    total count is already set error.
    """
    pass



# -*- coding: utf-8 -*-
"""
token manager module.
"""

import jwt

import pyrin.configuration.services as config_services

from pyrin.core.structs import Context, DTO, Manager
from pyrin.security.token import TokenPackage
from pyrin.security.token.interface import AbstractTokenBase
from pyrin.utils.custom_print import print_warning
from pyrin.security.token.exceptions import InvalidTokenHandlerTypeError, \
    DuplicatedTokenHandlerError, TokenHandlerNotFoundError, InvalidTokenHandlerNameError, \
    TokenKidHeaderNotSpecifiedError, TokenKidHeaderNotFoundError, DuplicatedTokenKidHeaderError, \
    TokenDecodingError


class TokenManager(Manager):
    """
    token manager class.
    """

    package_class = TokenPackage

    def __init__(self):
        """
        initializes an instance of TokenManager.
        """

        super().__init__()

        self._token_handlers = Context()

        # a dictionary containing the relation between each kid to the handler name.
        # in the form of: {str kid: str handler_name}
        self._kid_to_handler_map = DTO()

    def register_token_handler(self, instance, **options):
        """
        registers a new token handler or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding an instance which it's name is already available
        in registered handlers.

        :param AbstractTokenBase instance: token handler to be registered.
                                           it must be an instance of AbstractTokenBase.

        :keyword bool replace: specifies that if there is another registered
                               handler with the same name, replace it with
                               the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidTokenHandlerTypeError: invalid token handler type error.
        :raises InvalidTokenHandlerNameError: invalid token handler name error.
        :raises DuplicatedTokenHandlerError: duplicated token handler error.
        :raises DuplicatedTokenKidHeaderError: duplicated token kid header error.
        """

        if not isinstance(instance, AbstractTokenBase):
            raise InvalidTokenHandlerTypeError('Input parameter [{instance}] is '
                                               'not an instance of [{base}].'
                                               .format(instance=instance,
                                                       base=AbstractTokenBase))

        if instance.get_name() is None or len(instance.get_name().strip()) == 0:
            raise InvalidTokenHandlerNameError('Token handler [{instance}] does '
                                               'not have a valid name.'
                                               .format(instance=instance))

        # checking whether is there any registered instance with the same name.
        if instance.get_name() in self._token_handlers.keys():
            replace = options.get('replace', False)

            if replace is not True:
                raise DuplicatedTokenHandlerError('There is another registered token handler '
                                                  'with name [{name}] but "replace" option is '
                                                  'not set, so handler [{instance}] '
                                                  'could not be registered.'
                                                  .format(name=instance.get_name(),
                                                          instance=instance))

            old_instance = self._token_handlers[instance.get_name()]
            print_warning('Token handler [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance, new_instance=instance))

        # checking whether is there any registered instance with the same kid header.
        if instance.get_kid() in self._kid_to_handler_map.keys():
            raise DuplicatedTokenKidHeaderError('There is another registered token handler '
                                                'with "kid" header [{kid}]. each handler '
                                                'must have a unique "kid" header.'
                                                .format(kid=instance.get_kid()))

        # registering new token handler and it's mapping.
        self._token_handlers[instance.get_name()] = instance
        self._kid_to_handler_map[instance.get_kid()] = instance.get_name()

    def _get_token_handler(self, **options):
        """
        gets the specified token handler.

        :keyword str handler_name: name of token handler to get.

        :raises TokenHandlerNotFoundError: token handler not found error.

        :rtype: AbstractTokenBase
        """

        handler_name = options.get('handler_name', self._get_default_handler_name())
        if handler_name not in self._token_handlers:
            raise TokenHandlerNotFoundError('Token handler [{name}] not found.'
                                            .format(name=handler_name))

        return self._token_handlers[handler_name]

    def generate_access_token(self, payload, **options):
        """
        generates an access token using specified handler from the given inputs and returns it.

        the generated token is in the form of `header_hash.payload_hash.signature_hash`
        and each part is encoded using a signing key.

        :param dict payload: a dictionary containing key/values as payload.
                             note that for better performance, keep payload
                             as small as possible.

        :keyword str handler_name: name of token handler to be used.
                                   if not provided, default handler
                                   from relevant configs will be used.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :keyword bool is_fresh: indicates that this token is fresh.
                                being fresh means that token is generated by
                                providing user credentials to server.
                                if not provided, defaults to False.

        :raises TokenHandlerNotFoundError: token handler not found error.

        :returns: token.
        :rtype: str
        """

        return self._get_token_handler(**options).generate_access_token(payload, **options)

    def generate_refresh_token(self, payload, **options):
        """
        generates a refresh token using specified handler from the given inputs and returns it.

        the generated token is in the form of `header_hash.payload_hash.signature_hash`
        and each part is encoded using a signing key.

        :param dict payload: a dictionary containing key/values as payload.
                             note that for better performance, keep payload
                             as small as possible.

        :keyword str handler_name: name of token handler to be used.
                                   if not provided, default handler
                                   from relevant configs will be used.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :raises TokenHandlerNotFoundError: token handler not found error.

        :returns: token.
        :rtype: str
        """

        return self._get_token_handler(**options).generate_refresh_token(payload, **options)

    def get_payload(self, token, **options):
        """
        decodes token using correct handler and gets the payload data.

        :param str token: token to get it's payload.

        :raises TokenKidHeaderNotSpecifiedError: token kid header not specified error.
        :raises TokenKidHeaderNotFoundError: token kid header not found error.
        :raises TokenHandlerNotFoundError: token handler not found error.
        :raises TokenVerificationError: token verification error.

        :rtype: dict
        """

        handler_name = self._get_handler_name(token)
        return self._get_token_handler(handler_name=handler_name).get_payload(token, **options)

    def get_unverified_payload(self, token, **options):
        """
        decodes token and gets the payload data without verifying the signature.

        note that returned payload must not be trusted for any critical operations.

        :param str token: token to get it's payload.

        :raises TokenKidHeaderNotSpecifiedError: token kid header not specified error.
        :raises TokenKidHeaderNotFoundError: token kid header not found error.
        :raises TokenHandlerNotFoundError: token handler not found error.
        :raises TokenDecodingError: token decoding error.

        :rtype: dict
        """

        handler_name = self._get_handler_name(token)
        return self._get_token_handler(handler_name=handler_name).get_unverified_payload(
            token, **options)

    def get_unverified_header(self, token, **options):
        """
        gets the header dict of token without verifying the signature.

        note that the returned header must not be trusted for critical operations.

        :param str token: token to get it's header.

        :raises TokenDecodingError: token decoding error.

        :rtype: dict
        """

        try:
            return jwt.get_unverified_header(token)
        except Exception as error:
            raise TokenDecodingError(error) from error

    def generate_key(self, handler_name, **options):
        """
        generates a valid key for the given handler and returns it.

        :param str handler_name: token handler name to be used.

        :keyword int length: the length of generated key in bytes.
                             note that some token handlers may not accept custom
                             key length so this value would be ignored on those handlers.

        :rtype: str | tuple[str, str]
        """

        return self._get_token_handler(handler_name=handler_name).generate_key(**options)

    def _get_handler_name(self, token):
        """
        gets the handler name for specified token using the kid header.

        :param str token: token to get it's handler name.

        :raises TokenDecodingError: token decoding error.
        :raises TokenKidHeaderNotSpecifiedError: token kid header not specified error.
        :raises TokenKidHeaderNotFoundError: token kid header not found error.

        :rtype: str
        """

        header = self.get_unverified_header(token)
        if 'kid' not in header.keys():
            raise TokenKidHeaderNotSpecifiedError('The "kid" header must be present in token.')

        kid = header['kid']
        if kid not in self._kid_to_handler_map.keys():
            raise TokenKidHeaderNotFoundError('Token kid header [{kid}] not found.'
                                              .format(kid=kid))

        return self._kid_to_handler_map[kid]

    def _get_default_handler_name(self):
        """
        gets default token handler name from configs.

        :rtype: str
        """

        return config_services.get('security', 'token', 'default_token_handler')



# -*- coding: utf-8 -*-
"""
caching hooks module.
"""

import pyrin.caching.services as caching_services

from pyrin.application.decorators import application_hook
from pyrin.application.enumerations import ApplicationStatusEnum
from pyrin.application.hooks import ApplicationHookBase


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def application_status_changed(self, old_status, new_status):
        """
        this method will be called whenever application status changes.

        :param str old_status: old application status.
        :param str new_status: new application status.

        :enum status:
            INITIALIZING = 'Initializing'
            LOADING = 'Loading'
            READY = 'Ready'
            RUNNING = 'Running'
            TERMINATED = 'Terminated'
        """

        if new_status == ApplicationStatusEnum.TERMINATED:
            caching_services.persist_all(clear=True)

    def prepare_runtime_data(self):
        """
        this method will be get called after application has been fully initialized.

        any changes of this method to database, will be committed automatically by
        application. so you should not commit anything in the hook. if you do commit
        manually, unexpected behaviors may occur.

        note that this method will not get called when application starts in scripting mode.
        """

        caching_services.load_all()

    def after_application_loaded(self):
        """
        this method will be called after application has been loaded.
        """

        caching_services.clear_required_caches()



# -*- coding: utf-8 -*-
"""
authorization services module.
"""

from pyrin.application.services import get_component
from pyrin.security.authorization import AuthorizationPackage


def authorize(user, permissions, **options):
    """
    authorizes the given user for specified permissions.
    if user does not have each one of specified permissions,
    an error will be raised.

    :param user: user identity to authorize permissions for.

    :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                              for user authorization.

    :raises UserNotAuthenticatedError: user not authenticated error.
    :raises AuthorizationFailedError: authorization failed error.
    """

    return get_component(AuthorizationPackage.COMPONENT_NAME).authorize(user,
                                                                        permissions,
                                                                        **options)


def is_authorized(permissions, user=None, **options):
    """
    gets a value indicating that specified user is authorized for given permissions.

    :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                              for authorization.

    :param user: user identity to be checked for authorization.
                 if not provided, current user will be used.

    :rtype: bool
    """

    return get_component(AuthorizationPackage.COMPONENT_NAME).is_authorized(permissions,
                                                                            user,
                                                                            **options)



# -*- coding: utf-8 -*-
"""
packaging enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class PackageScopeEnum(CoreEnum):
    """
    package scope enum.
    """

    PYRIN = 0
    EXTENDED_APPLICATION = 1
    OTHER_APPLICATION = 2
    CUSTOM_APPLICATION = 3
    TEST = 4
    EXTENDED_UNIT_TEST = 5
    OTHER_UNIT_TEST = 6
    EXTENDED_INTEGRATION_TEST = 7
    OTHER_INTEGRATION_TEST = 8
    UNKNOWN = 100



# -*- coding: utf-8 -*-
"""
deserializer handlers datetime module.
"""

import pyrin.globalization.datetime.services as datetime_services

from pyrin.core.globals import NULL
from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase
from pyrin.converters.deserializer.decorators import deserializer
from pyrin.utils.datetime import DEFAULT_DATE_TIME_ISO_REGEX, DEFAULT_DATE_ISO_REGEX, \
    DEFAULT_TIME_ISO_REGEX, DEFAULT_LOCAL_NAIVE_TIME_REGEX, DEFAULT_UTC_ZULU_DATE_TIME_REGEX, \
    DEFAULT_LOCAL_NAIVE_DATE_TIME_REGEX


@deserializer()
class DateDeserializer(StringPatternDeserializerBase):
    """
    date deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of DateDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for date
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.
        but if any error occurs, it will be raised.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: date
        """

        converted_date = datetime_services.to_date(value)
        if converted_date is not None:
            return converted_date

        return NULL

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(DEFAULT_DATE_ISO_REGEX, 10, 10)]


@deserializer()
class TimeDeserializer(StringPatternDeserializerBase):
    """
    time deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of TimeDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for time
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.
        but if any error occurs, it will be raised.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: time
        """

        converted_time = datetime_services.to_time(value)
        if converted_time is not None:
            return converted_time

        return NULL

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(DEFAULT_TIME_ISO_REGEX, 14, 21),
                (DEFAULT_LOCAL_NAIVE_TIME_REGEX, 8, 15)]


@deserializer()
class DateTimeDeserializer(StringPatternDeserializerBase):
    """
    datetime deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of DateTimeDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for datetime
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.
        but if any error occurs, it will be raised.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: datetime
        """

        matching_pattern = options.get('matching_pattern')
        replace_server = None
        if matching_pattern == DEFAULT_LOCAL_NAIVE_DATE_TIME_REGEX:
            replace_server = False

        converted_datetime = datetime_services.to_datetime(value, server=False,
                                                           replace_server=replace_server)
        if converted_datetime is not None:
            return converted_datetime

        return NULL

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(DEFAULT_DATE_TIME_ISO_REGEX, 25, 32),
                (DEFAULT_UTC_ZULU_DATE_TIME_REGEX, 20, 27),
                (DEFAULT_LOCAL_NAIVE_DATE_TIME_REGEX, 19, 26)]



# -*- coding: utf-8 -*-
"""
mimetype handlers base module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.processor.mimetype.enumerations import MIMETypeEnum
from pyrin.processor.mimetype.exceptions import InvalidMIMETypeHandlerTypeError
from pyrin.processor.mimetype.interface import AbstractMIMETypeHandlerBase


class MIMETypeHandlerBase(AbstractMIMETypeHandlerBase):
    """
    mimetype handler base class.
    """

    def __init__(self, **options):
        """
        initializes an instance of MIMETypeHandlerBase.
        """

        super().__init__()
        self._next_handler = None

    def mimetype(self, value, **options):
        """
        gets the mimetype of given value.

        returns None if it fails to detect the mimetype.

        :param object value: value to detect its mimetype.

        :rtype: str
        """

        mimetype = None
        if self.is_valid(value, **options) is True:
            mimetype = self._mimetype(value, **options)

        if mimetype is None:
            if self._next_handler is not None:
                return self._next_handler.mimetype(value, **options)
            elif isinstance(value, str):
                return MIMETypeEnum.TEXT

        return mimetype

    @abstractmethod
    def _mimetype(self, value, **options):
        """
        gets the mimetype of given value.

        returns None if it fails to detect the mimetype.

        :param object value: value to detect its mimetype.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def set_next(self, mimetype_handler):
        """
        sets the next mimetype handler and returns it.

        :param AbstractMIMETypeHandlerBase mimetype_handler: mimetype handler instance
                                                             to be set as next handler.

        :rtype: AbstractMIMETypeHandlerBase
        """

        if mimetype_handler is not None and not isinstance(mimetype_handler,
                                                           MIMETypeHandlerBase):
            raise InvalidMIMETypeHandlerTypeError('Input parameter [{instance}] is '
                                                  'not an instance of [{base}].'
                                                  .format(instance=mimetype_handler,
                                                          base=MIMETypeHandlerBase))

        self._next_handler = mimetype_handler
        return mimetype_handler

    def is_valid(self, value, **options):
        """
        gets a value indicating that the given value's type is valid for this handler.

        :param object value: value to be determined for validity.

        :rtype: bool
        """

        return isinstance(value, self.accepted_type)



# -*- coding: utf-8 -*-
"""
string normalizer exceptions module.
"""

from pyrin.core.exceptions import CoreException


class StringNormalizerManagerException(CoreException):
    """
    string normalizer manager exception.
    """
    pass


class InvalidStringNormalizerTypeError(StringNormalizerManagerException):
    """
    invalid string normalizer type error.
    """
    pass


class DuplicatedStringNormalizerError(StringNormalizerManagerException):
    """
    duplicated string normalizer error.
    """
    pass


class StringNormalizerDoesNotExistError(StringNormalizerManagerException):
    """
    string normalizer does not exist error.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm types custom module.
"""

import uuid

from sqlalchemy.types import CHAR
from sqlalchemy.dialects.mssql import UNIQUEIDENTIFIER
from sqlalchemy.dialects.postgresql import UUID

from pyrin.database.orm.types.base import CoreCustomType


class GUID(CoreCustomType):
    """
    guid type class.

    it's a platform-independent guid type.
    uses postgresql's UUID type or mssql's UNIQUEIDENTIFIER
    type, otherwise uses CHAR(36), storing as stringified hex values.
    """

    impl = CHAR

    def load_dialect_impl(self, dialect):
        """
        returns a `TypeEngine` object corresponding to a dialect.

        :param Dialect dialect: the dialect in use.

        :rtype: TypeEngine
        """

        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        elif dialect.name == 'mssql':
            return dialect.type_descriptor(UNIQUEIDENTIFIER())
        else:
            return dialect.type_descriptor(CHAR(36))

    def process_bind_param(self, value, dialect):
        """
        receive a bound parameter value to be converted.

        :param TypeEngine value: data to operate upon. it could be `None`.
        :param Dialect dialect: the dialect in use.

        :rtype: str
        """

        if value is None:
            return value
        elif dialect.name in ('postgresql', 'mssql'):
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return str(uuid.UUID(value))
            else:
                return str(value)

    def process_result_value(self, value, dialect):
        """
        receive a result-row column value to be converted.

        :param str value: data to operate upon. it could be `None`.
        :param Dialect dialect: the dialect in use.

        :rtype: uuid.UUID
        """

        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value

    def compare_against_backend(self, dialect, conn_type):
        """
        returns True if this type is the same as the given database type,
        or None to allow the default implementation to compare these
        types. a return value of False means the given type does not
        match this type.

        :param Dialect dialect: the dialect in use.
        :param TypeEngine conn_type: type of the returned value from database.

        :rtype: bool
        """

        if dialect.name == 'postgresql':
            return isinstance(conn_type, UUID)
        elif dialect.name == 'mssql':
            return isinstance(conn_type, UNIQUEIDENTIFIER)
        else:
            return isinstance(conn_type, CHAR)



# -*- coding: utf-8 -*-
"""
configuration package.
"""

from pyrin.packaging.base import Package


class ConfigurationPackage(Package):
    """
    configuration package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.converters.deserializer']
    COMPONENT_NAME = 'configuration.component'



# -*- coding: utf-8 -*-
"""
database paging services module.
"""

from pyrin.application.services import get_component
from pyrin.database.paging import DatabasePagingPackage


def extract_paging_params(values):
    """
    extracts paging parameters from given dict and returns them as a new dict.

    the values will be removed from input dict.

    :param dict values: a dict to extract paging params from it.

    :returns: dict(int page: page number,
                   int page_size: page size)
    :rtype: dict
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).extract_paging_params(values)


def get_paging_params(**options):
    """
    gets paging parameters from given inputs.

    note that this method does not do any validation and just returns
    keys as they are, even if their value is None.

    :keyword int page: page number.
    :keyword int page_size: page size.

    :returns: tuple[int page, int page_size]
    :rtype: tuple[int, int]
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).get_paging_params(**options)


def generate_paging_params(page, page_size):
    """
    generates paging parameters from given inputs.

    :param int page: page number.
    :param int page_size: page size.

    :returns: dict[int page, int page_size]
    :rtype: dict[int, int]
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).generate_paging_params(page,
                                                                                      page_size)


def inject_paging_keys(limit, offset, values):
    """
    injects paging keys into given dict.

    :param int limit: limit.
    :param int offset: offset.
    :param dict values: a dict to inject paging keys into it.
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).inject_paging_keys(
        limit, offset, values)


def get_paging_keys(**options):
    """
    gets paging keys from given inputs.

    note that this method does not do any validation and just returns
    keys as they are, even if their value is None.

    :keyword int __limit__: limit value.
    :keyword int __offset__: offset value.

    :returns: tuple[int limit, int offset]
    :rtype: tuple[int, int]
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).get_paging_keys(**options)


def disable_paging_keys(values):
    """
    disables paging keys in given dict.

    :param dict values: a dict to disable paging keys in it.
    """

    return get_component(DatabasePagingPackage.COMPONENT_NAME).disable_paging_keys(values)



# -*- coding: utf-8 -*-
"""
caching exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CachingManagerException(CoreException):
    """
    caching manager exception.
    """
    pass


class CachingManagerBusinessException(CoreBusinessException, CachingManagerException):
    """
    caching manager business exception.
    """
    pass


class KeyIsNotPresentInCacheError(CachingManagerException):
    """
    key is not present in cache error.
    """
    pass


class InvalidCacheTypeError(CachingManagerException):
    """
    invalid cache type error.
    """
    pass


class DuplicatedCacheError(CachingManagerException):
    """
    duplicated cache error.
    """
    pass


class CacheNotFoundError(CachingManagerException):
    """
    cache not found error.
    """
    pass


class CacheIsNotPersistentError(CachingManagerException):
    """
    cache is not persistent error.
    """
    pass


class CacheNameIsRequiredError(CachingManagerException):
    """
    cache name is required error.
    """
    pass


class InvalidCacheLimitError(CachingManagerException):
    """
    invalid cache limit error.
    """
    pass


class InvalidCacheExpireTimeError(CachingManagerException):
    """
    invalid cache expire time error.
    """
    pass



# -*- coding: utf-8 -*-
"""
request wrappers package.
"""

from pyrin.packaging.base import Package


class RequestWrappersPackage(Package):
    """
    request wrappers package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
caching local handlers complex module.
"""

from threading import Lock

from pyrin.caching.local.containers.dict import OrderedDictContainer
from pyrin.caching.decorators import cache
from pyrin.caching.local.handlers.base import ComplexLocalCacheBase
from pyrin.caching.local.items.complex import ComplexLocalCacheItem


@cache()
class ComplexLocalCache(ComplexLocalCacheBase):
    """
    complex local cache class.

    this type of caches support expire time and count limit for items.
    it does also consider method inputs, current user and component key
    for cache key generation.
    """

    cache_name = 'complex'
    container_class = OrderedDictContainer
    cache_item_class = ComplexLocalCacheItem
    clearance_lock_class = Lock
    persistent_lock_class = Lock



# -*- coding: utf-8 -*-
"""
users component module.
"""

from pyrin.application.decorators import component
from pyrin.security.users import UsersPackage
from pyrin.security.users.manager import UsersManager
from pyrin.application.structs import Component


@component(UsersPackage.COMPONENT_NAME)
class UsersComponent(Component, UsersManager):
    """
    users component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
rsa256 encryption handler module.
"""

from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes

import pyrin.configuration.services as config_services
import pyrin.security.utils.services as security_utils_services

from pyrin.security.encryption.decorators import encrypter
from pyrin.security.encryption.handlers.base import RSAEncrypterBase


@encrypter()
class RSA256Encrypter(RSAEncrypterBase):
    """
    rsa256 encrypter class.
    """

    def __init__(self, **options):
        """
        initializes an instance of RSA256Encrypter.
        """

        super().__init__(**options)

    def _get_algorithm(self, **options):
        """
        gets the algorithm used for encryption.

        :rtype: str
        """

        return 'RSA256'

    def _encrypt(self, text, **options):
        """
        encrypts the given value and returns the encrypted result.

        :param str text: text to be encrypted.

        :rtype: bytes
        """

        return self._public_key.encrypt(text.encode(self._encoding),
                                        padding.OAEP(
                                            mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                            algorithm=hashes.SHA256(),
                                            label=None))

    def _decrypt(self, value, **options):
        """
        decrypts the given value and returns the decrypted result.

        :param bytes value: value to be decrypted.

        :rtype: str
        """

        return self._private_key.decrypt(value, padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None)).decode(self._encoding)

    def generate_key(self, **options):
        """
        generates a valid public/private key for this handler and returns it.

        :returns: tuple[str public_key, str private_key]
        :rtype: tuple[str, str]
        """

        return super().generate_key(length=2048)

    def _load_keys(self, **options):
        """
        loads public/private keys into this class's relevant attributes.
        """

        public_pem = config_services.get('security', 'encryption', 'rsa256_public_key')
        private_pem = config_services.get('security', 'encryption', 'rsa256_private_key')

        self._public_key, self._private_key = security_utils_services.load_rsa_key(public_pem,
                                                                                   private_pem,
                                                                                   **options)



# -*- coding: utf-8 -*-
"""
schema services module.
"""

from pyrin.application.services import get_component
from pyrin.api.schema import SchemaPackage


def get_computed_entity_columns(entity, result_schema=None, **options):
    """
    gets a dict containing all computed columns to be added to the result.

    if `result_schema` is not provided, it returns an empty dict.
    note that the result dict should not contain any `BaseEntity` or
    `ROW_RESULT` values, otherwise a max recursion error may occur.

    :param BaseEntity entity: the actual entity to be processed.
    :param ResultSchema result_schema: result schema to be used for computation.

    :rtype: dict
    """

    return get_component(SchemaPackage.COMPONENT_NAME).get_computed_entity_columns(entity,
                                                                                   result_schema,
                                                                                   **options)


def get_computed_row_columns(row, result_schema=None, **options):
    """
    gets a dict containing all computed columns to be added to the result.

    if `result_schema` is not provided, it returns an empty dict.
    note that the result dict should not contain any `BaseEntity` or
    `ROW_RESULT` values, otherwise a max recursion error may occur.

    :param ROW_RESULT row: the actual row result to be processed.
    :param ResultSchema result_schema: result schema to be used for computation.

    :rtype: dict
    """

    return get_component(SchemaPackage.COMPONENT_NAME).get_computed_row_columns(row,
                                                                                result_schema,
                                                                                **options)



# -*- coding: utf-8 -*-
"""
token handlers base module.
"""

import time

from abc import abstractmethod

import jwt

import pyrin.configuration.services as config_services
import pyrin.security.utils.services as security_utils_services

from pyrin.core.structs import DTO
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.enumerations import TokenTypeEnum
from pyrin.security.token.exceptions import TokenVerificationError, TokenDecodingError
from pyrin.security.token.interface import AbstractTokenBase
from pyrin.settings.static import APPLICATION_ENCODING
from pyrin.utils import unique_id


class TokenBase(AbstractTokenBase):
    """
    token base class.
    """

    def __init__(self, name, **options):
        """
        initializes an instance of TokenBase.

        :param str name: token handler name.
        """

        super().__init__()

        self._set_name(name)
        self._encoding = APPLICATION_ENCODING

    def generate_access_token(self, payload, **options):
        """
        generates an access token from the given inputs and returns it.

        the generated token is in the form of `header_hash.payload_hash.signature_hash`
        and each part is encoded using a signing key.

        :param dict payload: a dictionary containing key/values as payload.
                             note that for better performance, keep payload
                             as small as possible.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :keyword bool is_fresh: indicates that this token is fresh.
                                being fresh means that token is generated by
                                providing user credentials to server.
                                if not provided, defaults to False.

        :returns: token.
        :rtype: str
        """

        updated_payload = payload or DTO()
        updated_payload.update(**self._get_access_token_required_claims())

        return self._generate_token(updated_payload, **options)

    def generate_refresh_token(self, payload, **options):
        """
        generates a refresh token from the given inputs and returns it.

        the generated token is in the form of `header_hash.payload_hash.signature_hash`
        and each part is encoded using a signing key.

        :param dict payload: a dictionary containing key/values as payload.
                             note that for better performance, keep payload
                             as small as possible.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :returns: token.
        :rtype: str
        """

        updated_payload = payload or DTO()
        updated_payload.update(**self._get_refresh_token_required_claims())

        # refresh tokens never could be fresh.
        updated_payload.update(is_fresh=False)

        return self._generate_token(updated_payload, **options)

    def _generate_token(self, payload, **options):
        """
        generates a token from the given inputs and returns it.

        :param dict payload: a dictionary containing key/values as payload.
                             note that for better performance, keep payload
                             as small as possible.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :keyword bool is_fresh: indicates that this token is fresh.
                                being fresh means that token is generated by
                                providing user credentials to server.
                                if not provided, defaults to False.

        :returns: token.
        :rtype: str
        """

        is_fresh = options.get('is_fresh', False)
        updated_payload = payload or DTO()
        updated_payload.update(is_fresh=is_fresh)

        custom_headers = options.get('custom_headers', DTO())
        custom_headers.update(**self._get_common_required_headers())

        return jwt.encode(updated_payload,
                          self._get_encoding_key(),
                          self._get_algorithm(),
                          custom_headers)

    def get_payload(self, token, **options):
        """
        decodes token and gets the payload data.

        if token signature could not be verified, it will raise an error.

        :param str token: token to get it's payload.

        :raises TokenVerificationError: token verification error.

        :rtype: dict
        """

        try:
            return self._get_payload(token, verify=True)
        except Exception as error:
            raise TokenVerificationError(error) from error

    def _get_payload(self, token, **options):
        """
        decodes token and gets the payload data.

        :param str token: token to get it's payload.

        :keyword bool verify: specifies that token must be verified before getting
                              the payload. if not provided, defaults to True.

        :rtype: dict
        """

        default_options = self._get_default_options()
        verify = options.get('verify', True)
        if verify is False:
            default_options = dict(verify_signature=False)

        return jwt.decode(token,
                          self._get_decoding_key(),
                          self._get_algorithm(**options),
                          default_options)

    def get_unverified_payload(self, token, **options):
        """
        decodes token and gets the payload data without verifying the signature.

        note that returned payload must not be trusted for any critical operations.

        :param str token: token to get it's payload.

        :raises TokenDecodingError: token decoding error.

        :rtype: dict
        """

        try:
            return self._get_payload(token, verify=False)
        except Exception as error:
            raise TokenDecodingError(error) from error

    def get_unverified_header(self, token, **options):
        """
        gets the header dict of token without verifying the signature.

        note that returned header must not be trusted for critical operations.

        :param str token: token to get it's header.

        :rtype: dict
        """

        return jwt.get_unverified_header(token)

    def _get_common_required_headers(self):
        """
        gets a dictionary containing common required headers.

        :returns: dict[str kid: str kid]
        :rtype: dict
        """

        return DTO(kid=self.get_kid())

    def _get_common_required_claims(self):
        """
        gets a dictionary containing common required claims for access and refresh tokens.

        :returns: dict(str jti: jwt id,
                       float iat: issued time)

        :rtype: dict
        """

        now = time.time()

        return DTO(jti=str(unique_id.generate_uuid4()),
                   iat=now)

    def _get_access_token_required_claims(self):
        """
        gets a dictionary containing required claims for access token.

        :returns: dict(str jti: jwt id,
                       float iat: issued time,
                       float exp: expire time)

        :rtype: dict
        """

        common_required = self._get_common_required_claims()
        expire_duration = self._get_access_token_lifetime()
        expiration = common_required['iat'] + expire_duration
        common_required.update(exp=expiration,
                               type=TokenTypeEnum.ACCESS)

        return common_required

    def _get_refresh_token_required_claims(self):
        """
        gets a dictionary containing required claims for refresh token.

        :returns: dict(str jti: jwt id,
                       float iat: issued time,
                       float exp: expire time)

        :rtype: dict
        """

        common_required = self._get_common_required_claims()
        expire_duration = self._get_refresh_token_lifetime()
        expiration = common_required['iat'] + expire_duration
        common_required.update(exp=expiration,
                               type=TokenTypeEnum.REFRESH)

        return common_required

    def _get_default_options(self):
        """
        gets a dictionary containing default options.

        :returns: dict(bool verify_signature: verify signature,
                       bool verify_exp: verify expire time,
                       bool verify_nbf: verify not before,
                       bool verify_iat: verify issued at,
                       bool verify_aud: verify audience,
                       bool verify_iss: verify issuer,
                       bool require_exp: requires expire time,
                       bool require_iat: requires issued time,
                       bool require_nbf: requires not before time)

        :rtype: dict
        """

        return DTO(verify_signature=True,
                   verify_exp=True,
                   verify_nbf=True,
                   verify_iat=True,
                   verify_aud=True,
                   verify_iss=True,
                   require_exp=True,
                   require_iat=True,
                   require_nbf=False)

    @abstractmethod
    def _get_encoding_key(self, **options):
        """
        gets the signing key for encoding.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_decoding_key(self, **options):
        """
        gets the signing key for decoding.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def _get_access_token_lifetime(self, **options):
        """
        gets access token lifetime.

        :rtype: int
        """

        return config_services.get('security', 'token', 'access_token_lifetime', **options)

    def _get_refresh_token_lifetime(self, **options):
        """
        gets refresh token lifetime.

        :rtype: int
        """

        return config_services.get('security', 'token', 'refresh_token_lifetime', **options)

    @abstractmethod
    def _get_algorithm(self, **options):
        """
        gets the algorithm for signing the token.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()


class SymmetricTokenBase(TokenBase):
    """
    symmetric token base class.

    this token type uses a single symmetric key for decoding and encoding.
    """

    def __init__(self, **options):
        """
        initializes an instance of SymmetricTokenBase.
        """

        # we pass the algorithm of token handler as the name of it.
        super().__init__(self._get_algorithm(**options), **options)

    def _get_decoding_key(self, **options):
        """
        gets the signing key for decoding.

        :rtype: str
        """

        return self._get_encoding_key(**options)


class AsymmetricTokenBase(TokenBase):
    """
    asymmetric token base class.

    this token type uses a pair of public/private asymmetric keys for decoding and encoding.
    """

    def __init__(self, **options):
        """
        initializes an instance of AsymmetricTokenBase.
        """

        # we pass the algorithm of token handler as the name of it.
        super().__init__(self._get_algorithm(**options), **options)


class RSTokenBase(AsymmetricTokenBase):
    """
    rs token base class.

    this token type uses a pair of public/private asymmetric keys for decoding and encoding.
    """

    def __init__(self, **options):
        """
        initializes an instance of AsymmetricTokenBase.
        """

        super().__init__(**options)

        self._private_key = None
        self._public_key = None

        self._load_keys(**options)

    def generate_key(self, **options):
        """
        generates a valid public/private key for this handler and returns it.

        :keyword int length: the length of generated key in bits.
                             if not provided, it uses default value
                             from relevant config.

        :returns: tuple[str public_key, str private_key]
        :rtype: tuple[str, str]
        """

        return security_utils_services.generate_rsa_key(**options)

    @abstractmethod
    def _load_keys(self, **options):
        """
        loads public/private keys into this class's relevant attributes.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
string normalizer package.
"""

from pyrin.packaging.base import Package


class StringNormalizerPackage(Package):
    """
    string normalizer package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'utilities.string.normalizer.component'



# -*- coding: utf-8 -*-
"""
alembic enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class AlembicCLIHandlersEnum(CoreEnum):
    """
    alembic cli handlers enum.
    """

    ENABLE = 'alembic.enable'
    BRANCHES = 'branches'
    CURRENT = 'current'
    DOWNGRADE = 'downgrade'
    HEADS = 'heads'
    HISTORY = 'history'
    MERGE = 'merge'
    REVISION = 'revision'
    SHOW = 'show'
    STAMP = 'stamp'
    UPGRADE = 'upgrade'



# -*- coding: utf-8 -*-
"""
pbkdf2 hashing handler module.
"""

import hashlib
import re

import pyrin.configuration.services as config_services
import pyrin.security.utils.services as security_utils_services

from pyrin.security.hashing.decorators import hashing
from pyrin.security.hashing.handlers.base import HashingBase
from pyrin.security.hashing.handlers.exceptions import InvalidHashingRoundsCountError, \
    InvalidPBKDF2InternalAlgorithmError, InvalidHashingSaltLengthError


@hashing()
class PBKDF2Hashing(HashingBase):
    """
    pbkdf2 hashing class.
    """

    # regular expression to validate format of full hashed values.
    # the following format will be matched:
    # `$handler_name$internal_algorithm$rounds$salt_length$salt-text_plus_salt_hash`
    FORMAT_REGEX = re.compile(r'^\$PBKDF2\$[^$]+\$[\d]+\$[\d]+\$(.+)$')

    def __init__(self, **options):
        """
        initializes an instance of PBKDF2Hashing.
        """

        super().__init__(**options)

    def _generate_hash(self, text, **options):
        """
        gets the hash of input text using a random or specified salt.

        :param str text: text to be hashed.

        :keyword bytes salt: salt to be used for hashing.
                             if not provided, a random salt will be generated
                             considering `salt_length` option.

        :keyword str internal_algorithm: internal algorithm to be used
                                         for hashing. if not provided,
                                         default value from relevant
                                         config will be used.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :keyword int salt_length: salt length to be used for hashing.
                                  if `salt` option is provided, then
                                  this value will be ignored.
                                  if not provided, default value from
                                  relevant config will be used.

        :rtype: bytes
        """

        internal_algorithm, rounds, salt_length = self._extract_attributes(**options)
        self._validate_attributes(internal_algorithm, rounds, salt_length)

        salt = options.get('salt', None)
        if salt is None:
            salt = self._generate_salt(length=salt_length)

        text_hash = hashlib.pbkdf2_hmac(internal_algorithm,
                                        text.encode(self._encoding),
                                        salt,
                                        rounds)

        return self._make_final_hash(internal_algorithm, rounds, salt, text_hash)

    def _generate_salt(self, **options):
        """
        generates a valid salt for this handler and returns it.

        :keyword int length: length of generated salt in bytes.
                             if not provided, default value from
                             relevant config will be used.

        :rtype: bytes
        """

        salt_length = options.get('length', config_services.get('security', 'hashing',
                                                                'pbkdf2_salt_length'))

        return security_utils_services.get_bytes(length=salt_length)

    def _is_match(self, text, hashed_value, **options):
        """
        gets a value indicating that given text's
        hash is identical to given hashed value.

        :param str text: text to be hashed.
        :param bytes hashed_value: hashed value to compare with.

        :rtype: bool
        """

        internal_algorithm, rounds, salt, text_hash = \
            self._extract_parts_from_final_hash(hashed_value, **options)

        new_full_hashed_value = self._generate_hash(text,
                                                    internal_algorithm=internal_algorithm,
                                                    rounds=rounds, salt=salt)

        return hashed_value == new_full_hashed_value

    def _get_algorithm(self, **options):
        """
        gets the hashing algorithm.

        :rtype: str
        """

        return 'PBKDF2'

    def _get_separator_count(self):
        """
        gets the separator count used between parts of this handler's hashed result.

        :rtype: int
        """

        return 5

    def _extract_attributes(self, **options):
        """
        extracts the required attributes for this handler from input
        keyword arguments. if not available, gets the default
        values from relevant configs.

        :keyword str internal_algorithm: internal algorithm to be used
                                         for hashing. if not provided,
                                         default value from relevant
                                         config will be used.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :keyword int salt_length: salt length to be used for hashing.
                                  if not provided, default value from
                                  relevant config will be used.

        :returns: tuple[str internal_algorithm, int rounds, int salt_length]
        :rtype: tuple[str, int, int]
        """

        internal_algorithm = options.get('internal_algorithm',
                                         config_services.get('security', 'hashing',
                                                             'pbkdf2_internal_algorithm'))

        rounds = options.get('rounds', config_services.get('security', 'hashing',
                                                           'pbkdf2_rounds'))

        salt_length = options.get('salt_length', config_services.get('security', 'hashing',
                                                                     'pbkdf2_salt_length'))

        return internal_algorithm, rounds, salt_length

    def _validate_attributes(self, internal_algorithm, rounds, salt_length):
        """
        validates the given inputs for hash generation.
        it will raise an error on invalid inputs.

        :param str internal_algorithm: internal algorithm to be used for hashing.
        :param int rounds: rounds to perform for generating hash.
        :param int salt_length: salt length to be used for hashing.

        :raises InvalidPBKDF2InternalAlgorithmError: invalid pbkdf2 internal algorithm error.
        :raises InvalidHashingRoundsCountError: invalid hashing rounds count error.
        :raises InvalidHashingSaltLengthError: invalid hashing salt length error.
        """

        if internal_algorithm not in hashlib.algorithms_guaranteed:
            raise InvalidPBKDF2InternalAlgorithmError('Internal algorithm [{algorithm}] '
                                                      'is invalid.'
                                                      .format(algorithm=internal_algorithm))

        if rounds < 1:
            raise InvalidHashingRoundsCountError('Hashing rounds [{rounds}] is invalid.'
                                                 .format(rounds=rounds))

        if salt_length < 1:
            raise InvalidHashingSaltLengthError('Salt length [{length}] is invalid.'
                                                .format(length=salt_length))

    def _make_final_hash(self, internal_algorithm, rounds, salt, text_hash):
        """
        makes final hash from input values and returns it.

        :param str internal_algorithm: internal algorithm to be used for hashing.
        :param int rounds: rounds to perform for generating hash.
        :param bytes salt: salt to be used for hashing.
        :param bytes text_hash: hash value of text and salt.

        :rtype: bytes
        """

        return self._get_separator() + self._get_separator().join(
            (self._get_algorithm().encode(self._encoding),
             internal_algorithm.encode(self._encoding),
             str(rounds).encode(self._encoding),
             str(len(salt)).encode(self._encoding),
             self._encode_hash_part(salt + text_hash)))

    def _extract_parts_from_final_hash(self, full_hashed_value, **options):
        """
        extracts different parts of given full hashed value.

        :param bytes full_hashed_value: full hashed value to extract it's parts.

        :returns: tuple[str internal_algorithm, int rounds, bytes salt, bytes text_hash]
        :rtype: tuple[str, int, bytes, bytes]
        """

        empty, handler, internal_algorithm, rounds, salt_length, salt_plus_text_hash = \
            full_hashed_value.split(self._get_separator(), self._get_separator_count())

        salt_length = int(salt_length)
        raw_salt_plus_text_hash = self._decode_hash_part(salt_plus_text_hash)
        salt = raw_salt_plus_text_hash[:salt_length]
        text_hash = raw_salt_plus_text_hash[salt_length:]

        return internal_algorithm.decode(self._encoding), int(rounds), salt, text_hash

    def _get_hashed_part(self, full_hashed_value, **options):
        """
        gets the hashed part from full hashed value which current handler understands it.
        this handler returns the same input value as result.

        :param bytes full_hashed_value: full hashed value to get hashed part from it.

        :rtype: bytes
        """

        return full_hashed_value



# -*- coding: utf-8 -*-
"""
utils encoding module.
"""

import base64

from pyrin.settings.static import APPLICATION_ENCODING


def bytes_to_base64(value):
    """
    encodes the given raw bytes into base64 encoded bytes.

    :param bytes value: raw bytes value to be encoded into base64 bytes.

    :returns: base64 encoded bytes.

    :rtype: bytes
    """

    return base64.b64encode(value)


def base64_to_bytes(value):
    """
    decodes the given base64 encoded bytes value into raw bytes.

    :param bytes value: base64 encoded bytes value to be decoded into raw bytes.

    :returns: raw bytes.

    :rtype: bytes
    """

    return base64.b64decode(value)


def bytes_to_base64_string(value):
    """
    encodes the given raw bytes into base64 encoded bytes string.

    :param bytes value: raw bytes value to be encoded into base64 bytes string.

    :returns: base64 encoded bytes string.

    :rtype: str
    """

    return bytes_to_base64(value).decode(APPLICATION_ENCODING)


def base64_string_to_bytes(value):
    """
    decodes the given base64 encoded bytes string value into raw bytes.

    :param str value: base64 encoded bytes string value to be decoded into raw bytes.

    :returns: raw bytes.

    :rtype: bytes
    """

    return base64_to_bytes(value.encode(APPLICATION_ENCODING))



# -*- coding: utf-8 -*-
"""
model manager module.
"""

from pyrin.core.structs import Manager
from pyrin.database.model import ModelPackage
from pyrin.database.model.base import CoreEntity


class ModelManager(Manager):
    """
    model manager class.
    """

    package_class = ModelPackage

    def get_declarative_base(self):
        """
        gets the application's declarative base class.

        by default it returns `CoreEntity` class.
        if you want to define a new declarative base class in your application
        instead of using `CoreEntity` which is provided by pyrin, you must
        override this method inside your application's `database.model.ModelManager`
        class and return your new declarative base class.
        this is required if you want to enable migrations in your application.
        note that your application must have a unique declarative base type and
        you shouldn't mix the usage of `CoreEntity` and your new declarative base,
        otherwise migrations will not work properly.

        but, if you don't want to use migrations at all, you could just put
        `pyrin.database.migration` into `ignored_packages` list of `packaging.ini`
        file and leave this method unimplemented.

        :rtype: type[CoreEntity]
        """

        return CoreEntity



# -*- coding: utf-8 -*-
"""
template decorators module.
"""

import pyrin.template.services as template_services


def template_handler(**options):
    """
    decorator to register a template handler.

    :keyword bool replace: specifies that if there is another registered
                           template handler with the same name, replace it
                           with the new one, otherwise raise
                           an error. defaults to False.

    :raises InvalidTemplateHandlerTypeError: invalid template handler type error.
    :raises DuplicatedTemplateHandlerError: duplicated template handler error.

    :returns: template handler class.
    :rtype: AbstractTemplateHandler
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available template handlers.

        :param AbstractTemplateHandler cls: template handler class.

        :returns: template handler class.
        :rtype: AbstractTemplateHandler
        """

        instance = cls()
        template_services.register_template_handler(instance, **options)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
logging manager module.
"""

import logging
import logging.config

from logging import Logger

import pyrin.configuration.services as config_services

from pyrin.core.mixin import HookMixin
from pyrin.logging import LoggingPackage
from pyrin.core.structs import Manager
from pyrin.logging.adapters import RequestInfoLoggerAdapter, BaseLoggerAdapter
from pyrin.logging.hooks import LoggingHookBase
from pyrin.logging.exceptions import InvalidLoggerAdapterTypeError, LoggerNotExistedError, \
    InvalidLoggingHookTypeError


class LoggingManager(Manager, HookMixin):
    """
    logging manager class.
    """

    # the logger adapter could be overridden in subclasses.
    logger_adapter_class = RequestInfoLoggerAdapter
    hook_type = LoggingHookBase
    invalid_hook_type_error = InvalidLoggingHookTypeError
    package_class = LoggingPackage

    def __init__(self):
        """
        initializes an instance of LoggingManager.
        """

        super().__init__()

        self._config_file_path = config_services.get_file_path(
            self.package_class.EXTRA_CONFIG_STORE_NAMES[0])
        self._load_configs(self._config_file_path)

    def _load_configs(self, config_file_path):
        """
        loads logging configuration and handlers from given file.

        :param str config_file_path: config file path.
        """

        logging.config.fileConfig(config_file_path, disable_existing_loggers=True)

    def _wrap_root_logger(self):
        """
        wraps the root logger into an adapter.
        """

        logging.root = self._wrap_logger(logging.root)

    def _get_all_loggers(self):
        """
        gets a dictionary containing all available loggers.

        :returns: dict[str name: Logger instance]
        :rtype: dict
        """

        return logging.root.manager.loggerDict

    def _wrap_logger(self, logger):
        """
        wraps the given logger into an adapter and returns it.

        :param Logger logger: logger instance to be wrapped.

        :rtype: BaseLoggerAdapter | Logger
        """

        if self.should_be_wrapped(logger) is True:
            return self.logger_adapter_class(logger)

        return logger

    def _wrap_all_loggers(self):
        """
        wraps all available loggers into an adapter to inject request info into every log.

        note that we should not wrap sqlalchemy and alembic loggers,
        because it does not affect on sqlalchemy loggers and they
        never have request info in emitted logs, I don't know the reason.
        but wrapping them actually has a side effect which leads to some errors.
        so we do not wrap them in the first place.
        """

        self._wrap_root_logger()

        for name, logger in self._get_all_loggers().items():
            if self.should_be_wrapped(logger) is True:
                self._set_logger_adapter(name, self._wrap_logger(logger))

    def should_be_wrapped(self, logger):
        """
        gets a value indication that given logger should be wrapped.

        note that we should not wrap sqlalchemy and alembic loggers,
        because it does not affect on sqlalchemy loggers and they
        never have request info in emitted logs, I don't know the reason.
        but wrapping them actually has a side effect which leads to some errors.
        so we do not wrap them in the first place.

        :param Logger logger: logger to check should it be wrapped.

        :rtype: bool
        """

        if not isinstance(logger, Logger):
            return False

        unwrapped_loggers = config_services.get_active('logging', 'unwrapped_loggers')
        return all(item not in logger.name for item in unwrapped_loggers)

    def wrap_all_loggers(self):
        """
        wraps all available loggers into an adapter.

        normally, this method should not be called manually.
        """

        self._wrap_all_loggers()

    def _set_logger_adapter(self, name, adapter):
        """
        sets the logger adapter instance with the specified name into logging loggerDict.

        :param str name: name of the logger adapter.
        :param BaseLoggerAdapter adapter: logger adapter instance.

        :raises InvalidLoggerAdapterTypeError: invalid logger adapter type error.
        :raises LoggerNotExistedError: logger not existed error.
        """

        if not isinstance(adapter, BaseLoggerAdapter):
            raise InvalidLoggerAdapterTypeError('Input parameter [{adapter}] is not an '
                                                'instance of [{base}].'
                                                .format(adapter=adapter,
                                                        base=BaseLoggerAdapter))

        loggers = self._get_all_loggers()
        if name not in loggers:
            raise LoggerNotExistedError('Logger [{name}] does not exist.'
                                        .format(name=name))

        loggers[name] = adapter

    def get_all_loggers(self):
        """
        gets a dictionary containing all available loggers.

        it returns a shallow copy of loggers dict.

        :returns: dict[str name: Logger instance]
        :rtype: dict
        """

        return self._get_all_loggers().copy()

    def reload_configs(self, **options):
        """
        reloads all logging configurations from config file.
        """

        self._load_configs(self._config_file_path)

    def get_logger(self, name, **options):
        """
        gets the logger based on input parameters.

        :param str name: logger name to get.

        :returns: specified logger.
        :rtype: BaseLoggerAdapter
        """

        logger = logging.getLogger(name)
        if self.should_be_wrapped(logger) is True:
            logger = self._wrap_logger(logger)
            self._set_logger_adapter(name, logger)

        return logger

    def debug(self, msg, *args, **kwargs):
        """
        emits a log with debug level.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        logging.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        emits a log with info level.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        logging.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        emits a log with warning level.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        logging.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        emits a log with error level.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        logging.error(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        """
        emits a log with error level and exception information.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        logging.exception(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        emits a log with critical level.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        logging.critical(msg, *args, **kwargs)

    def prepare_data(self, data, **options):
        """
        this method will call `prepare_data` method of all registered hooks.

        :param dict | object data: data that is passed to logging method.

        :returns: modified or same input data.
        :rtype: dict | object
        """

        prepared_data = data
        for hook in self._get_hooks():
            prepared_data = hook.prepare_data(prepared_data, **options)

        return prepared_data

    def before_emit(self, message, data, level, **options):
        """
        this method will call `before_emit` method of all registered hooks.

        :param str message: the log message that must be emitted.
        :param dict | object data: data that is passed to logging method.
        :param int level: log level.
        """

        for hook in self._get_hooks():
            hook.before_emit(message, data, level, **options)

    def after_emit(self, message, data, level, **options):
        """
        this method will call `after_emit` method of all registered hooks.

        :param str message: the log message that has been emitted.
        :param dict | object data: data that is passed to logging method.
        :param int level: log level.
        """

        for hook in self._get_hooks():
            hook.after_emit(message, data, level, **options)



# -*- coding: utf-8 -*-
"""
caching mixin package.
"""

from pyrin.packaging.base import Package


class CachingMixinPackage(Package):
    """
    caching mixin package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
schema package.
"""

from pyrin.packaging.base import Package


class SchemaPackage(Package):
    """
    schema package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.configuration']
    COMPONENT_NAME = 'api.schema.component'



# -*- coding: utf-8 -*-
"""
logging exceptions module.
"""

from pyrin.core.exceptions import CoreException


class LoggingManagerException(CoreException):
    """
    logging manager exception.
    """
    pass


class InvalidLoggerAdapterTypeError(LoggingManagerException):
    """
    invalid logger adapter type error.
    """
    pass


class LoggerNotExistedError(LoggingManagerException):
    """
    logger not existed error.
    """
    pass


class InvalidLoggingHookTypeError(LoggingManagerException):
    """
    invalid logging hook type error.
    """
    pass



# -*- coding: utf-8 -*-
"""
serializer handlers dictionary module.
"""

import pyrin.converters.serializer.services as serializer_services

from pyrin.core.structs import DTO
from pyrin.converters.serializer.decorators import serializer
from pyrin.converters.serializer.handlers.base import SerializerBase


@serializer()
class DictionarySerializer(SerializerBase):
    """
    dictionary serializer class.
    """

    def _serialize(self, value, **options):
        """
        serializes the given value.

        :param dict value: dict value to be serialized.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :rtype: dict
        """

        result = DTO(value)
        for key, item in result.items():
            result[key] = serializer_services.serialize(item, **options)

        return result

    @property
    def accepted_type(self):
        """
        gets the accepted type for this serializer.

        which could serialize values from this type.

        :rtype: type[dict]
        """

        return dict



# -*- coding: utf-8 -*-
"""
database hooks module.
"""

import pyrin.database.services as database_services

from pyrin.application.decorators import application_hook
from pyrin.application.hooks import ApplicationHookBase
from pyrin.core.structs import Hook
from pyrin.packaging.decorators import packaging_hook
from pyrin.packaging.hooks import PackagingHookBase


class DatabaseHookBase(Hook):
    """
    database hook base class.
    """

    def after_session_factories_configured(self):
        """
        this method will be called after all database session factories have been configured.
        """
        pass


@packaging_hook()
class PackagingHook(PackagingHookBase):
    """
    packaging hook class.
    """

    def after_packages_loaded(self):
        """
        this method will be called after all application packages have been loaded.
        """

        # we have to configure session factories after all models have
        # been loaded to enable multiple database connections if needed.
        database_services.configure_session_factories()


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def finalize_transaction(self, response, **options):
        """
        this method will be called after a request has been fully processed.

        :param CoreResponse response: response object.

        :rtype: CoreResponse
        """

        return database_services.finalize_transaction(response, **options)



# -*- coding: utf-8 -*-
"""
alembic handlers stamp module.
"""

from pyrin.database.migration.alembic.enumerations import AlembicCLIHandlersEnum
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.handlers.params import SQLParam, TagParam, \
    PurgeParam, RevisionsParam


@alembic_cli_handler()
class StampCLIHandler(AlembicCLIHandlerBase):
    """
    stamp cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of StampCLIHandler.
        """

        super().__init__(AlembicCLIHandlersEnum.STAMP)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([SQLParam(), TagParam(),
                       PurgeParam(), RevisionsParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
pyrin

an application framework to facilitate developing
small to large scale applications based on flask
which are more maintainable and extensible.
"""


__version__ = '0.4.16'



# -*- coding: utf-8 -*-
"""
logging masking manager module.
"""

from werkzeug.datastructures import Headers

import pyrin.configuration.services as config_services

from pyrin.core.structs import Manager
from pyrin.logging.masking import LoggingMaskingPackage
from pyrin.logging.masking.structs import MaskedDict


class LoggingMaskingManager(Manager):
    """
    logging masking manager class.
    """

    # these are default keywords that will be masked if they are present in a dict.
    # these values could be extended or disabled in 'logging.masking' config store.
    # note that all values will be normalized and matched against also normalized values.
    # for example 'api_key' and 'APIkey' and 'api-key' will be considered equal.
    DEFAULT_MASKED_KEYWORDS = {'access_token',
                               'address',
                               'api_key',
                               'api_token',
                               'auth_token',
                               'authentication',
                               'authorization',
                               'birth_certificate',
                               'birth_certificate_no',
                               'birth_certificate_number',
                               'birth_date',
                               'birthday',
                               'card_no',
                               'card_number',
                               'certificate',
                               'cookie',
                               'cookies',
                               'credential',
                               'credentials',
                               'credit_card',
                               'credit_card_no',
                               'credit_card_number',
                               'cvv',
                               'debit_card',
                               'debit_card_no',
                               'debit_card_number',
                               'email',
                               'email_address',
                               'fingerprint',
                               'hash',
                               'hash_value',
                               'home_address',
                               'http_auth',
                               'http_authorization',
                               'http_cookie',
                               'identity_code',
                               'identity_no',
                               'identity_number',
                               'mobile',
                               'mobile_no',
                               'mobile_number',
                               'national_code',
                               'national_no',
                               'national_number',
                               'oauth',
                               'oauth_token',
                               'otp',
                               'pass',
                               'passport',
                               'passport_no',
                               'passport_number',
                               'passwd',
                               'password',
                               'password_hash',
                               'phone',
                               'phone_no',
                               'phone_number',
                               'postal_address',
                               'postal_code',
                               'private_key',
                               'proxy_authorization',
                               'refresh_token',
                               'rsa',
                               'rsa_key',
                               'secret',
                               'secret_key',
                               'session',
                               'session_id',
                               'token',
                               'uploaded_files',
                               'username',
                               'www_authenticate',
                               'zip_code'}

    # a class type to be used as masked dict.
    masked_dict_class = MaskedDict
    package_class = LoggingMaskingPackage

    def __init__(self):
        """
        initializes an instance of LoggingMaskingManager.
        """

        super().__init__()

        self._masks = self._get_masked_keywords()

    def _get_masked_keywords(self):
        """
        gets all keywords that should be masked.

        :rtype: set[str]
        """

        normalized_defaults = self._normalize_items(self.DEFAULT_MASKED_KEYWORDS)
        include_defaults = config_services.get_active('logging.masking',
                                                      'include_default_keywords')
        extra_keywords = config_services.get_active('logging.masking', 'extra_keywords')
        extra_keywords = self._normalize_items(extra_keywords)

        results = list(extra_keywords)
        if include_defaults is True:
            results.extend(normalized_defaults)

        return set(results)

    def _normalize_items(self, items):
        """
        normalizes all given items and returns a new list.

        :param list[str] | tuple[str] | set[str] items: items to be normalized.

        :rtype: list[str]
        """

        results = []
        for item in items:
            results.append(self._normalize(item))

        return results

    def _normalize(self, value):
        """
        normalizes given value.

        if the value is not a string, it returns the value itself.
        this method does not check `isinstance` to detect strings, because it
        has a significant performance hit. so it returns the same value
        whenever an `AttributeError` is occurred.

        :param str | object value: value to be normalized.

        :rtype: str | object
        """

        try:
            return value.lower().replace('_', '').replace('-', '') \
                .replace('.', '').replace(' ', '')
        except AttributeError:
            return value

    def should_mask(self, key):
        """
        gets a value indicating that given key should be masked in a dict.

        :param str key: dict key to be checked.

        :rtype: bool
        """

        return self._normalize(key) in self._masks

    def mask(self, data, **options):
        """
        masks the given data if required.

        if data is not a dict, no changes will be made.

        :param dict | object data: data to be masked.

        :returns: a masked dict or original input data.
        :rtype: MaskedDict | object
        """

        if (isinstance(data, dict) and not
           isinstance(data, self.masked_dict_class)) or isinstance(data, Headers):
            return self.masked_dict_class(data)

        return data



# -*- coding: utf-8 -*-
"""
babel component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.globalization.locale.babel import BabelPackage
from pyrin.globalization.locale.babel.manager import BabelManager


@component(BabelPackage.COMPONENT_NAME)
class BabelComponent(Component, BabelManager):
    """
    babel component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
deserializer handlers tuple module.
"""

import re

import pyrin.converters.deserializer.services as deserializer_services

from pyrin.converters.deserializer.decorators import deserializer
from pyrin.converters.deserializer.handlers.base import DeserializerBase, \
    StringPatternDeserializerBase


@deserializer()
class TupleDeserializer(DeserializerBase):
    """
    tuple deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of TupleDeserializer.

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes every possible value available in input tuple.

        gets a new deserialized tuple.

        :param tuple value: value that should be deserialized.

        :rtype: tuple
        """

        result_list = []
        for item in value:
            result_list.append(deserializer_services.deserialize(item, **options))

        return tuple(result_list)

    @property
    def accepted_type(self):
        """
        gets the accepted type for this deserializer.

        which could deserialize values from this type.

        :rtype: type[tuple]
        """

        return tuple


@deserializer()
class StringTupleDeserializer(StringPatternDeserializerBase):
    """
    string tuple deserializer class.

    note that this deserializer could only handle tuples with single depth.
    meaning that nested tuples are not supported. and also nested lists or
    dictionaries or sets or any other collections are not supported and
    stops deserialization.
    for example: (1, (2, 4), [5, 4]) will not be deserialized.
    """

    # default min for this deserializer is 2 because
    # it should at least have ( and ) at both ends.
    DEFAULT_MIN = 2

    # matches a tuple inside string, all of these values will be matched.
    # example: (), (1), (1,), (1,2), (1,2,)
    # it won't accept nested collections, all of these values won't match.
    # example: ([]), (1, {2: 4}, (2,3))
    TUPLE_REGEX = re.compile(r'^\(\)$|^\([^\(\){}\[\]]+(,[^\(\){}\[\]]+)*\)$')

    def __init__(self, **options):
        """
        creates an instance of StringTupleDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  patterns and their min and
                                                                  max length for tuple
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: tuple
        """

        # removing the first '(' and last ')' from value.
        value = value[1:-1]
        items = value.split(',')
        temp_list = []
        for item in items:
            if len(item.strip()) > 0:
                temp_list.append(item.strip())

        return deserializer_services.deserialize(tuple(temp_list), **options)

    @property
    def default_formats(self):
        """
        gets default accepted patterns that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.TUPLE_REGEX, self.UNDEF_LENGTH, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
authorization manager module.
"""

import pyrin.security.services as security_services
import pyrin.security.users.services as user_services
import pyrin.security.session.services as session_services
import pyrin.utils.misc as misc_utils

from pyrin.core.globals import _
from pyrin.core.structs import Manager
from pyrin.security.authorization import AuthorizationPackage
from pyrin.security.authorization.exceptions import AuthorizationFailedError, \
    UserNotAuthenticatedError, UserIsNotActiveError


class AuthorizationManager(Manager):
    """
    authorization manager class.

    this class is intended to provide some services needed in pyrin application.
    the top level application must extend this class considering business requirements.
    """

    package_class = AuthorizationPackage

    def authorize(self, user, permissions, **options):
        """
        authorizes the given user for specified permissions.

        if user does not have each one of specified permissions,
        an error will be raised.

        :param user: user identity to authorize permissions for.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for user authorization.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        if user is None:
            raise UserNotAuthenticatedError(_('User has not been authenticated.'))

        if user_services.is_active(user, **options) is not True:
            message = _('User [{user}] is not active.')
            raise UserIsNotActiveError(message.format(user=user))

        # we must check whether input permissions object is iterable.
        # if not, we make it manually.
        required_permissions = permissions
        required_permissions = misc_utils.make_iterable(required_permissions, tuple)

        if len(required_permissions) > 0:
            if security_services.has_permission(user, required_permissions,
                                                **options) is not True:
                message = _('User [{user}] does not have the '
                            'required permission(s) {permissions}.')
                raise AuthorizationFailedError(
                    message.format(user=user,
                                   permissions=list(required_permissions)))

    def is_authorized(self, permissions, user=None, **options):
        """
        gets a value indicating that specified user is authorized for given permissions.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for authorization.

        :param user: user identity to be checked for authorization.
                     if not provided, current user will be used.

        :rtype: bool
        """

        if user is None:
            user = session_services.get_current_user()

        try:
            self.authorize(user, permissions)
            return True
        except AuthorizationFailedError:
            return False



# -*- coding: utf-8 -*-
"""
session package.
"""

from pyrin.packaging.base import Package


class SessionPackage(Package):
    """
    session package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'security.session.component'



# -*- coding: utf-8 -*-
"""
celery cli exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CeleryCLIManagerException(CoreException):
    """
    celery cli manager exception.
    """
    pass


class CeleryCLIManagerBusinessException(CoreBusinessException,
                                        CeleryCLIManagerException):
    """
    celery cli manager business exception.
    """
    pass



# -*- coding: utf-8 -*-
"""
database package.
"""

from pyrin.packaging.base import Package


class DatabasePackage(Package):
    """
    database package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.component'

    DEPENDS = ['pyrin.configuration',
               'pyrin.logging']

    CONFIG_STORE_NAMES = ['database',
                          'database.binds']



# -*- coding: utf-8 -*-
"""
alembic handlers upgrade module.
"""

from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.enumerations import AlembicCLIHandlersEnum
from pyrin.database.migration.alembic.handlers.base import \
    AlembicUpgradeDowngradeCLIHandlerBase


@alembic_cli_handler()
class UpgradeCLIHandler(AlembicUpgradeDowngradeCLIHandlerBase):
    """
    upgrade cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of UpgradeCLIHandler.
        """

        super().__init__(AlembicCLIHandlersEnum.UPGRADE)



# -*- coding: utf-8 -*-
"""
locale exceptions module.
"""

from pyrin.core.exceptions import CoreException


class LocaleManagerException(CoreException):
    """
    locale manager exception.
    """
    pass


class InvalidLocaleSelectorTypeError(LocaleManagerException):
    """
    invalid locale selector type error.
    """
    pass


class InvalidTimezoneSelectorTypeError(LocaleManagerException):
    """
    invalid timezone selector type error.
    """
    pass


class LocaleSelectorHasBeenAlreadySetError(LocaleManagerException):
    """
    locale selector has been already set error.
    """
    pass


class TimezoneSelectorHasBeenAlreadySetError(LocaleManagerException):
    """
    timezone selector has been already set error.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli core services module.
"""

from pyrin.cli.core.manager import CLICoreManager


def get_manager():
    """
    gets the cli core manager instance.

    :rtype: CLICoreManager
    """

    return CLICoreManager()


def create(handler_name):
    """
    creates the required templates using relevant handlers.

    :param str handler_name: handler name to be used.

    :raises CLICoreTemplateHandlerNotFoundError: cli core template handler not found error.
    """

    return get_manager().create(handler_name)



# -*- coding: utf-8 -*-
"""
caching models module.
"""

from sqlalchemy import BigInteger, Unicode, LargeBinary

from pyrin.database.model.base import CoreEntity
from pyrin.database.model.mixin import HistoryMixin
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.orm.sql.schema.columns import AutoPKColumn


class CacheItemBaseEntity(CoreEntity):
    """
    cache item base entity class.
    """

    _table = 'cache_item'

    id = AutoPKColumn(name='id')


class CacheItemEntity(CacheItemBaseEntity, HistoryMixin):
    """
    cache item entity class.
    """

    _extend_existing = True

    cache_name = CoreColumn(name='cache_name', type_=Unicode(length=25), nullable=False)
    shard_name = CoreColumn(name='shard_name', type_=Unicode(25), nullable=True)
    version = CoreColumn(name='version', type_=Unicode(length=20), nullable=False)
    key = CoreColumn(name='key', type_=BigInteger, nullable=False)
    item = CoreColumn(name='item', type_=LargeBinary, nullable=False)



# -*- coding: utf-8 -*-
"""
cors component module.
"""

from pyrin.processor.cors import CORSPackage
from pyrin.processor.cors.manager import CORSManager
from pyrin.application.decorators import component
from pyrin.application.structs import Component


@component(CORSPackage.COMPONENT_NAME)
class CORSComponent(Component, CORSManager):
    """
    cors component class.
    """
    pass
